
''' A module for representing train tracks on triangulations. '''

import networkx

import curver
from curver.kernel.lamination import Shortenable  # Special import needed for subclassing.
from curver.kernel.utilities import memoize  # Special import needed for decorating.

class TrainTrack(Shortenable):
    ''' A Lamination in which each triangle is tripod free. '''
    
    def shorten_strategy(self, edge):
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        if not self.triangulation.is_flippable(edge):
            return 0
        
        a, b, c, d, e = self.triangulation.square(edge)
        ad, bd, cd, dd, ed = [self.dual_weight(edgy) for edgy in self.triangulation.square(edge)]
        if self(e) <= 0:
            return 0
        if b == ~d and ad > 0 and bd == 0 and ed == 0:
            return 0
        if a == ~c and ad == 0 and bd > 0 and ed == 0:
            return 0
        
        num_bad = [ed > 0, ad < 0, bd < 0].count(True)
        num_good = [ed < 0, ad > 0, bd > 0].count(True)
        
        if num_bad > 0:
            return 0.25
        
        if num_good == 3:
            return 1
        elif num_good == 2:
            return 1
        else:  # num_good == 1:
            if (ad > 0 and cd > 0) or (bd > 0 and dd > 0):
                return 0.5
            else:
                return 0.75
    
    @memoize()
    def components(self):
        short, conjugator = self.shorten()
        
        components = dict()
        for edge in short.triangulation.positive_edges:  # Only need to check half of them.
            # Check for an Arc here.
            if short(edge) < 0:
                component, multiplicity = short.triangulation.edge_arc(edge), abs(short(edge))
                components[conjugator.inverse()(component)] = multiplicity  # Map it back onto self.
            elif short(edge) > 0:
                if short.triangulation.is_flippable(edge):  # Check for a non-peripheral curve here.
                    a, b, c, d, e = short.triangulation.square(edge)
                    da, db, dc, dd, de = [short.dual_weight(edgy) for edgy in short.triangulation.square(edge)]
                    if b == ~d and db == 0 and de == 0:
                        geometric = [1 if index == b.index or index == e.index else 0 for index in short.triangulation.indices]
                        component, multiplicity = curver.kernel.Curve(short.triangulation, geometric), short(e)
                        components[conjugator.inverse()(component)] = multiplicity  # Map it back onto self.
                else:  # Check for a peripheral curve here.
                    geometric = [1 if index == edge.index else 0 for index in short.triangulation.indices]
                    component, multiplicity = curver.kernel.Curve(short.triangulation, geometric), short(edge)
                    components[conjugator.inverse()(component)] = multiplicity  # Map it back onto self.
        
        return components
    
    def vertex_cycles(self):
        ''' Yield the vertex cycles of this train track.
        
        These are the curves carried by this train track that run over each brach at most twice.
        Be careful as there are often a *lot* of them. '''
        
        def connected_to(edge):
            ''' Yield the edges you can reach by travelling out of the given edge. '''
            corner = self.triangulation.corner_lookup[edge.label]
            if self.dual_weight(corner[1]): yield ~corner[2]
            if self.dual_weight(corner[2]): yield ~corner[1]
        
        # Build graph.
        edges = [(edge, edgy) for edge in self.triangulation.edges for edgy in connected_to(edge)]
        G = networkx.DiGraph(edges)
        
        for cycle in networkx.simple_cycles(G):
            curve = self.triangulation.lamination_from_cut_sequence(cycle)
            if isinstance(curve, curver.kernel.Curve):
                yield curve
    
    def vertex_cycle(self):
        ''' Return a vertex cycle of this train track. '''
        
        return next(iter(self.vertex_cycles()))





''' A module for representing and manipulating maps between Triangulations. '''

from fractions import Fraction
import numpy as np

import curver

NT_TYPE_PERIODIC = 'Periodic'
NT_TYPE_REDUCIBLE = 'Reducible'  # Strictly this  means 'reducible and not periodic'.
NT_TYPE_PSEUDO_ANOSOV = 'Pseudo-Anosov'

class Encoding(object):
    ''' This represents a map between two Triagulations.
    
    The map is given by a sequence of Moves which act from right to left. '''
    def __init__(self, sequence):
        assert(isinstance(sequence, (list, tuple)))
        assert(sequence)
        # assert(all(isinstance(item, curver.kernel.Move) for item in sequence))  # Quadratic.
        
        self.sequence = sequence
        
        self.source_triangulation = self.sequence[-1].source_triangulation
        self.target_triangulation = self.sequence[0].target_triangulation
        self.zeta = self.source_triangulation.zeta
        if self.is_mapping_class():  # Promote.
            self.__class__ = MappingClass
    
    def is_mapping_class(self):
        ''' Return if this encoding is a mapping class.
        
        That is, if it maps to the triangulation it came from. '''
        
        return self.source_triangulation == self.target_triangulation
    
    def __repr__(self):
        return str(self)
    def __str__(self):
        return str(self.sequence)
    def __iter__(self):
        return iter(self.sequence)
    def __len__(self):
        return len(self.sequence)
    def __getitem__(self, value):
        if isinstance(value, slice):
            # It turns out that handling all slices correctly is really hard.
            # We need to be very careful with "empty" slices. As Encodings require
            # non-empty sequences, we have to return just the id_encoding. This
            # ensures the Encoding that we return satisfies:
            #   self == self[:i] * self[i:j] * self[j:]
            # even when i == j.
            
            start = 0 if value.start is None else value.start if value.start >= 0 else len(self) + value.start
            stop = len(self) if value.stop is None else value.stop if value.stop >= 0 else len(self) + value.stop
            if start == stop:
                if 0 <= start < len(self):
                    return self.sequence[start].target_triangulation.id_encoding()
                elif start == len(self):
                    return self.source_triangulation.id_encoding()
                else:
                    raise IndexError('list index out of range')
            elif stop < start:
                raise IndexError('list index out of range')
            else:  # start < stop.
                return Encoding(self.sequence[value])
        elif isinstance(value, curver.IntegerType):
            return self.sequence[value]
        else:
            return NotImplemented
    def package(self):
        ''' Return a small amount of info that self.source_triangulation can use to reconstruct this triangulation. '''
        return [item.package() for item in self]
    def __reduce__(self):
        return (create_encoding, (self.source_triangulation, self.package()))
    
    def __eq__(self, other):
        if isinstance(other, Encoding):
            if self.source_triangulation != other.source_triangulation or self.target_triangulation != other.target_triangulation:
                raise ValueError('Cannot compare Encodings between different triangulations.')
            
            return all(self(arc) == other(arc) for arc in self.source_triangulation.edge_arcs()) and \
                all(self(hc) == other(hc) for hc in self.source_triangulation.edge_homologies())  # We only really need this for S_{1,1}.
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    def __hash__(self):
        # In fact this hash is perfect unless the surface is S_{1,1}.
        return hash(tuple(entry for row in self.intersection_matrix().tolist() for entry in row))
    
    def __call__(self, other):
        if self.source_triangulation != other.triangulation:
            raise ValueError('Cannot apply an Encoding to something on a triangulation other than source_triangulation.')
        
        for item in reversed(self.sequence):
            other = item(other)
        
        return other
    def __mul__(self, other):
        if isinstance(other, Encoding):
            if self.source_triangulation != other.target_triangulation:
                raise ValueError('Cannot compose Encodings over different triangulations.')
            
            return Encoding(self.sequence + other.sequence)
        elif other is None:
            return self
        else:
            return NotImplemented
    def inverse(self):
        ''' Return the inverse of this encoding. '''
        
        return Encoding([item.inverse() for item in reversed(self.sequence)])
    def __invert__(self):
        return self.inverse()
    
    def homology_matrix(self):
        ''' Return a matrix describing the action of this encoding on first homology.
        
        The matrix is given with respect to the homology bases of the source and target triangulations. '''
        
        source_basis = self.source_triangulation.homology_basis()
        target_basis = self.target_triangulation.homology_basis()
        
        source_images = [self(hc).canonical() for hc in source_basis]
        
        return np.matrix([[sum(x * y for x, y in zip(hc, hc2)) for hc in source_images] for hc2 in target_basis])
    
    def intersection_matrix(self):
        ''' Return the matrix M = {signed_intersection(self(e_i), e'_j)}_{ij}.
        Here e_i and e'j are the edges of self.source_triangulation and self.target_triangulation respectively.
        
        Except when on S_{1,1}, this uniquely determines self. '''
        
        return np.matrix([list(self(arc)) for arc in self.source_triangulation.edge_arcs()])
    
    def vertex_map(self):
        ''' Return the dictionary (vertex, self(vertex)) for each vertex in self.source_triangulation.
        
        When self is a MappingClass this is a permutation of the vertices. '''
        
        source_vertices = dict((vertex, self.source_triangulation.curve_from_cut_sequence(vertex)) for vertex in self.source_triangulation.vertices)
        target_vertices_inverse = dict((self.target_triangulation.curve_from_cut_sequence(vertex), vertex) for vertex in self.target_triangulation.vertices)
        
        return dict((vertex, target_vertices_inverse[self(source_vertices[vertex])]) for vertex in self.source_triangulation.vertices)
    
    def simplify(self):
        ''' Return a new Encoding that is equal to self.
        
        This is obtained by combining the image of the source triangulation under self and so is (hopefully) simpler than self since it depends only on the endpoints. '''
        
        _, conjugator = self(self.source_triangulation.as_lamination()).shorten()
        # conjugator.inverse() is almost self, however the edge labels might not agree.
        potential_closers = [isom.encode() for isom in conjugator.target_triangulation.isometries_to(self.source_triangulation)]
        identity = self.source_triangulation.id_encoding()
        [closer] = [potential_closer for potential_closer in potential_closers if potential_closer * conjugator  * self == identity]  # Thre should only be one.
        
        return (closer * conjugator).inverse()

class MappingClass(Encoding):
    ''' An Encoding where self.source_triangulation == self.target_triangulation. '''
    def is_mapping_class(self):
        return True
    
    def __pow__(self, k):
        if k == 0:
            return self.source_triangulation.id_encoding()
        elif k > 0:
            return MappingClass(self.sequence * k)
        else:
            return self.inverse()**abs(k)
    
    def order(self):
        ''' Return the order of this mapping class.
        
        If this has infinite order then return 0. '''
        
        # There are several tricks that we could use to rule out powers that we need to test:
        #  - If self**i == identity then homology_matrix**i = identity_matrix.
        #  - If self**i == identity then self**(ij) == identity.
        #  - if self**i == identity and self**j == identity then self**gcd(i, j) == identity.
        #
        # But in terms of raw speed there doesn't appear to be anything faster than:
        
        homology_matrix = self.homology_matrix()
        originals = [np.identity(homology_matrix.shape[0]), self.source_triangulation.as_lamination()] + self.source_triangulation.edge_arcs()
        images = list(originals)
        powers = [0] * len(originals)
        for power in range(1, self.source_triangulation.max_order()+1):
            for i in range(len(originals)):
                while powers[i] < power:
                    images[i] = self(images[i]) if i > 0 else homology_matrix * images[i]
                    powers[i] += 1
                if (i == 0 and not np.array_equal(images[i], originals[i])) or (i > 0 and images[i] != originals[i]):
                    break
            else:
                return power
        
        return 0
    
    def is_identity(self):
        ''' Return if this mapping class is the identity. '''
        
        return self.order() == 1
    
    def is_periodic(self):
        ''' Return if this mapping class has finite order. '''
        
        return self.order() > 0
    
    def is_reducible(self):
        ''' Return if this mapping class is reducible (and *not* periodic). '''
        
        return self.nielsen_thurston_type() == NT_TYPE_REDUCIBLE
    
    def is_pseudo_anosov(self):
        ''' Return if this mapping class is pseudo-Anosov. '''
        
        return self.nielsen_thurston_type() == NT_TYPE_PSEUDO_ANOSOV
    
    def nielsen_thurston_type(self):
        ''' Return the Nielsen--Thurston type of this mapping class. '''
        
        if self.is_periodic():
            return NT_TYPE_PERIODIC
        elif self.positive_asymptotic_translation_length():
            return NT_TYPE_PSEUDO_ANOSOV
        else:  # self.asymptotic_translation_length() == 0:
            return NT_TYPE_REDUCIBLE
    
    def asymptotic_translation_length(self):
        ''' Return the asymptotic translation length of this mapping class on the curve complex.
        
        From Algorithm 6 of [BellWebb16]_. '''
        
        C = curver.kernel.CurveComplex(self.source_triangulation)
        c = self.source_triangulation.edge_arc(0).boundary()  # A "short" curve.
        geodesic = C.geodesic(c, (self**C.M)(c))
        m = geodesic[len(geodesic)//2]  # midpoint
        
        numerator = C.distance(m, (self**C.M)(m))
        denominator = C.M
        return Fraction(numerator, denominator).limit_denominator(C.D)
    
    def positive_asymptotic_translation_length(self):
        ''' Return whether the asymptotic translation length of this mapping class on the curve complex is positive
        
        This uses Remark 4.7 of [BellWebb16]_ and so is more efficient than doing::
        
            self.asymptotic_translation_length() > 0 '''
        
        C = curver.kernel.CurveComplex(self.source_triangulation)
        c = self.source_triangulation.edge_arc(0).boundary()  # A "short" curve.
        geodesic = C.geodesic(c, (self**C.M2)(c))
        m = geodesic[len(geodesic)//2]  # midpoint
        
        return C.distance(m, (self**C.M2)(m)) > 4


def create_encoding(source_triangulation, sequence):
    ''' Return the encoding defined by sequence starting at source_triangulation.
    
    This is only really here to help with pickling. Users should use
    source_triangulation.encode(sequence) directly. '''
    
    assert(isinstance(source_triangulation, curver.kernel.Triangulation))
    
    return source_triangulation.encode(sequence)





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies
import numpy as np

class TestEncoding(unittest.TestCase):
    def assertEqualArray(self, M, N):
        self.assertTrue(np.array_equal(M, N), msg='AssertionError: %s != %s' % (M, N))
    
    @given(strategies.encodings())
    def test_pickle(self, h):
        self.assertEqual(h, pickle.loads(pickle.dumps(h)))
    
    @given(st.data())
    def test_slice(self, data):
        h = data.draw(strategies.encodings())
        i = data.draw(st.integers(min_value=0, max_value=len(h)))
        j = data.draw(st.integers(min_value=0, max_value=len(h)))
        i, j = sorted([i, j])
        self.assertEqual(h[:i] * h[i:j] * h[j:], h)
    
    @given(st.data())
    @settings(max_examples=20)
    def test_inverse(self, data):
        g = data.draw(strategies.encodings())
        h = data.draw(strategies.encodings(g.target_triangulation))
        self.assertEqual(~(~g), g)
        self.assertEqual(~g * ~h, ~(h * g))
    
    @given(st.data())
    def test_homology_matrix(self, data):
        g = data.draw(strategies.encodings())
        h = data.draw(strategies.encodings(g.target_triangulation))
        self.assertEqualArray(h.homology_matrix() * g.homology_matrix(), (h * g).homology_matrix())
    
    @given(strategies.encodings())
    def test_intersection_matrix(self, h):
        self.assertEqualArray(h.intersection_matrix().transpose(), (~h).intersection_matrix())
    
    @given(strategies.encodings())
    def test_simplify(self, h):
        self.assertEqual(h.simplify(), h)
    
    @given(strategies.encodings())
    def test_vertex_map(self, h):
        vertex_map = h.vertex_map()
        self.assertEqual(sorted(vertex_map.keys()), sorted(h.source_triangulation.vertices))
        self.assertEqual(sorted(vertex_map.values()), sorted(h.target_triangulation.vertices))

class TestMappingClass(unittest.TestCase):
    @given(st.data())
    @settings(max_examples=20)
    def test_hash(self, data):
        mcg = data.draw(strategies.mcgs())
        g = data.draw(strategies.mapping_classes(mcg))
        h = data.draw(strategies.mapping_classes(mcg))
        self.assertTrue(hash(g) != hash(h) or g == h)
    
    @given(strategies.mapping_classes())
    # @settings(max_examples=2)
    def test_identity(self, h):
        self.assertEqual(h.is_identity(), h == h.source_triangulation.id_encoding())
    
    @given(strategies.mapping_classes())
    @settings(max_examples=2)
    def test_order(self, h):
        self.assertLessEqual(h.order(), h.source_triangulation.max_order())
        self.assertEqual(h**(h.order()), h.source_triangulation.id_encoding())





''' A module for representing basic ways of changing triangulations.
These moves can also track how laminations and homology classes move through those changes. '''

import curver

class Move(object):
    ''' A basic move from one triangulation to another. '''
    def __init__(self, source_triangulation, target_triangulation):
        assert(isinstance(source_triangulation, curver.kernel.Triangulation))
        assert(isinstance(target_triangulation, curver.kernel.Triangulation))
        
        self.source_triangulation = source_triangulation
        self.target_triangulation = target_triangulation
        self.zeta = self.source_triangulation.zeta
    def __repr__(self):
        return str(self)
    def __invert__(self):
        return self.inverse()
    def __call__(self, other):
        if isinstance(other, curver.kernel.Lamination):
            return self.apply_lamination(other)
        elif isinstance(other, curver.kernel.HomologyClass):
            return self.apply_homology(other)
        else:
            raise TypeError('Unknown type %s' % other)
    
    def encode(self):
        ''' Return the Encoding induced by this move. '''
        
        return curver.kernel.Encoding([self])
    
    def package(self):  #pylint: disable=no-self-use
        ''' Return a small amount of data such that self.source_triangulation.encode([data]) == self.encode(). '''
        
        return NotImplemented
    
    def inverse(self):  #pylint: disable=no-self-use
        ''' Return the inverse of this move. '''
        
        return NotImplemented
    def apply_lamination(self, lamination):  #pylint: disable=no-self-use,unused-argument
        ''' Return the lamination obtained by mapping the given lamination through this move. '''
        
        return NotImplemented
    def apply_homology(self, homology_class):  #pylint: disable=no-self-use,unused-argument
        ''' Return the homology class obtained by mapping the given homology class through this move. '''
        
        return NotImplemented

class Isometry(Move):
    ''' This represents an isometry from one Triangulation to another.
    
    Triangulations can create the isometries between themselves and this
    is the standard way users are expected to create these. '''
    def __init__(self, source_triangulation, target_triangulation, label_map):
        ''' This represents an isometry from source_triangulation to target_triangulation.
        
        It is given by a map taking each edge label of source_triangulation to a label of target_triangulation.
        
        Assumes that this map is defined on all labels. '''
        
        super(Isometry, self).__init__(source_triangulation, target_triangulation)
        
        assert(isinstance(label_map, dict))
        self.label_map = dict(label_map)
        
        # Quick sanity check.
        assert(all(i in self.label_map for i in self.source_triangulation.labels))
        
        self.index_map = dict((i, curver.kernel.norm(self.label_map[i])) for i in self.source_triangulation.indices)
        # Store the inverses too while we're at it.
        self.inverse_label_map = dict((self.label_map[label], label) for label in self.source_triangulation.labels)
        self.inverse_index_map = dict((index, curver.kernel.norm(self.inverse_label_map[index])) for index in self.source_triangulation.indices)
    
    def __str__(self):
        return 'Isometry ' + str([curver.kernel.Edge(self.label_map[index]) for index in self.source_triangulation.indices])
    def __reduce__(self):
        return (self.__class__, (self.source_triangulation, self.target_triangulation, self.label_map))
    def package(self):
        if not all(self.label_map[i] == i for i in self.source_triangulation.indices):  # If self is not the identity isometry.
            return {i: self.label_map[i] for i in self.source_triangulation.labels}
        else:
            return None
    def __eq__(self, other):
        if isinstance(other, Isometry):
            return self.source_triangulation == other.source_triangulation and self.target_triangulation == other.target_triangulation and self.label_map == other.label_map
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    
    def apply_lamination(self, lamination):
        geometric = [lamination(self.inverse_index_map[index]) for index in self.source_triangulation.indices]
        return lamination.__class__(self.target_triangulation, geometric)  # Avoids promote.
    
    def apply_homology(self, homology_class):
        algebraic = [homology_class(self.inverse_label_map[index]) for index in self.source_triangulation.indices]
        return curver.kernel.HomologyClass(self.target_triangulation, algebraic)
    
    def inverse(self):
        
        return Isometry(self.target_triangulation, self.source_triangulation, self.inverse_label_map)

class EdgeFlip(Move):
    ''' Represents the change to a curve caused by flipping an edge. '''
    def __init__(self, source_triangulation, target_triangulation, edge):
        super(EdgeFlip, self).__init__(source_triangulation, target_triangulation)
        
        if isinstance(edge, curver.IntegerType): edge = curver.kernel.Edge(edge)  # If given an integer instead.
        
        self.edge = edge
        assert(self.source_triangulation.is_flippable(self.edge))
        
        self.square = self.source_triangulation.square(self.edge)
    
    def __str__(self):
        return 'Flip %s' % self.edge
    def __reduce__(self):
        return (self.__class__, (self.source_triangulation, self.target_triangulation, self.edge))
    def package(self):
        return self.edge.label
    def __eq__(self, other):
        if isinstance(other, EdgeFlip):
            return self.source_triangulation == other.source_triangulation and self.target_triangulation == other.target_triangulation and self.edge == other.edge
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    
    def apply_lamination(self, lamination):
        L = lamination  # Shorter name.
        a, b, c, d, e = self.square
        ai, bi, ci, di, ei = [max(L(edge), 0) for edge in self.square]
        
        # Most of the new information matches the old, so we'll take a copy and modify the places that have changed.
        geometric = list(L.geometric)
        
        if L(e) >= ai + bi and ai >= di and bi >= ci:  # CASE: A(ab)
            geometric[e.index] = ai + bi - L(e)
        elif L(e) >= ci + di and di >= ai and ci >= bi:  # CASE: A(cd)
            geometric[e.index] = ci + di - L(e)
        elif L(e) <= 0 and ai >= bi and di >= ci:  # CASE: D(ad)
            geometric[e.index] = ai + di - L(e)
        elif L(e) <= 0 and bi >= ai and ci >= di:  # CASE: D(bc)
            geometric[e.index] = bi + ci - L(e)
        elif L(e) >= 0 and ai >= bi + L(e) and di >= ci + L(e):  # CASE: N(ad)
            geometric[e.index] = ai + di - 2*L(e)
        elif L(e) >= 0 and bi >= ai + L(e) and ci >= di + L(e):  # CASE: N(bc)
            geometric[e.index] = bi + ci - 2*L(e)
        elif ai + bi >= L(e) and bi + L(e) >= 2*ci + ai and ai + L(e) >= 2*di + bi:  # CASE: N(ab)
            geometric[e.index] = (ai + bi - L(e)) // 2
        elif ci + di >= L(e) and di + L(e) >= 2*ai + ci and ci + L(e) >= 2*bi + di:  # CASE: N(cd)
            geometric[e.index] = (ci + di - L(e)) // 2
        else:
            geometric[e.index] = max(ai + ci, bi + di) - L(e)
        
        return lamination.__class__(self.target_triangulation, geometric)  # Avoids promote.
    
    def apply_homology(self, homology_class):
        a, b, c, d, e = self.square
        
        algebraic = list(homology_class)
        # Move the homology on e onto a & b.
        algebraic[a.index] -= a.sign() * homology_class(e)
        algebraic[b.index] -= b.sign() * homology_class(e)
        algebraic[e.index] = 0
        
        return curver.kernel.HomologyClass(self.target_triangulation, algebraic)
    
    def inverse(self):
        return EdgeFlip(self.target_triangulation, self.source_triangulation, ~self.edge)





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies

class TestMultiCurve(unittest.TestCase):
    @given(strategies.multicurves())
    @settings(max_examples=20)
    def test_boundary_intersection(self, multicurve):
        boundary = multicurve.boundary()
        self.assertEqual(multicurve.intersection(boundary), 0)
    
    @given(st.data())
    @settings(max_examples=10)
    def test_boundary_union(self, data):
        multicurve = data.draw(strategies.multicurves().filter(lambda c: not c.is_peripheral()))  # Assume not peripheral.
        lamination = data.draw(strategies.laminations(multicurve.triangulation))
        boundary = multicurve.boundary_union(lamination)
        self.assertEqual(multicurve.intersection(boundary), 0)
        self.assertEqual(lamination.intersection(boundary), 0)
    
    @given(strategies.multicurves())
    @settings(max_examples=20)
    def test_crush(self, multicurve):
        crush = multicurve.crush()
        self.assertEqual(crush.source_triangulation.euler_characteristic, crush.target_triangulation.euler_characteristic)
    
    @given(strategies.multicurves())
    @settings(max_examples=20)
    def test_fills(self, multicurve):
        self.assertEqual(multicurve.is_filling(), multicurve.fills_with(multicurve))
    
    @given(strategies.multicurves())
    @settings(max_examples=10)
    def test_train_track(self, multicurve):
        train_track = multicurve.train_track()
        self.assertEqual(train_track.num_components(), multicurve.num_components())
        vertex_cycle = train_track.vertex_cycle()
        self.assertTrue(all(vertex_cycle.dual_weight(edge) == 0 or train_track.dual_weight(edge) > 0 for edge in train_track.triangulation.edges))
    
    @given(st.data())
    @settings(max_examples=2)
    def test_sum(self, data):
        triangulation = data.draw(strategies.triangulations())
        multicurves = data.draw(st.lists(elements=strategies.multicurves(triangulation), min_size=2, max_size=3))
        self.assertIsInstance(multicurves[0] + multicurves[1], curver.kernel.MultiCurve)
        self.assertIsInstance(triangulation.sum(multicurves), curver.kernel.MultiCurve)


class TestCurve(unittest.TestCase):
    def assertWithinOne(self, x, y):
        self.assertTrue(abs(x - y) <= 1, msg='AssertionError: |%s - %s| > 1' % (x, y))
    
    @given(strategies.curves())
    @settings(max_examples=50)
    def test_boundary_intersection(self, curve):
        boundary = curve.boundary()
        self.assertEqual(curve.intersection(boundary), 0)
    
    @given(st.data())
    @settings(max_examples=2)
    def test_sum(self, data):
        triangulation = data.draw(strategies.triangulations())
        multicurves = data.draw(st.lists(elements=strategies.multicurves(triangulation), min_size=2, max_size=3))
        self.assertIsInstance(multicurves[0] + multicurves[1], curver.kernel.MultiCurve)
        self.assertIsInstance(triangulation.sum(multicurves), curver.kernel.MultiCurve)
    
    @given(st.data())
    @settings(max_examples=10)
    def test_slope(self, data):
        curve = data.draw(strategies.curves().filter(lambda c: not c.is_peripheral()))  # Assume not peripheral.
        lamination = data.draw(strategies.laminations(curve.triangulation).filter(lambda l: curve.intersection(l) > 0))  # Assume intersect.
        slope = curve.slope(lamination)
        twist = curve.encode_twist()
        self.assertTrue(-1 <= slope <= 1 or curve.slope(twist(lamination)) == slope - 1)
    
    @given(st.data())
    @settings(max_examples=1)
    def test_relative_twisting(self, data):
        curve = data.draw(strategies.curves().filter(lambda c: not c.is_peripheral()))  # Assume not peripheral.
        lamination1 = data.draw(strategies.laminations(curve.triangulation).filter(lambda l: curve.intersection(l) > 0))  # Assume intersect.
        power = data.draw(st.integers())
        lamination2 = curve.encode_twist(power)(lamination1)
        rel_twisting = curve.relative_twisting(lamination1, lamination2)
        self.assertWithinOne(rel_twisting, power)
        
        h = data.draw(strategies.encodings(curve.triangulation))
        self.assertWithinOne(h(curve).relative_twisting(h(lamination1), h(lamination2)), power)
    
    @given(st.data())
    @settings(max_examples=20)
    def test_topological_type(self, data):
        curve = data.draw(strategies.curves())
        h = data.draw(strategies.encodings(curve.triangulation))
        self.assertEqual(curve.topological_type(), h(curve).topological_type())





''' The main window of the curver GUI application. '''

import curver
import curver.application

import re
import os
import io
import sys
import pickle
from math import sin, cos, pi, ceil, sqrt
from random import random
from colorsys import hls_to_rgb
from itertools import combinations
from collections import namedtuple

try:
    import Tkinter as TK
    import tkFileDialog
    import tkMessageBox
except ImportError:  # Python 3.
    try:
        import tkinter as TK
        import tkinter.filedialog as tkFileDialog
        import tkinter.messagebox as tkMessageBox
    except ImportError:
        raise ImportError('Tkinter not available.')

try:
    import ttk as TTK
except ImportError:  # Python 3.
    try:
        from tkinter import ttk as TTK
    except ImportError:
        raise ImportError('Ttk not available.')

# Some constants.
if sys.platform in ['darwin']:
    COMMAND = {
        'close': 'Command+W',
        }
    COMMAND_KEY = {
        'close': '<Command-w>',
        }
else:
    COMMAND = {
        'close': 'Ctrl+W',
        }
    COMMAND_KEY = {
        'close': '<Control-w>',
        }

# Vectors to offset a label by to produce backing.
OFFSETS = [(1.5*cos(2 * pi * i / 12), 1.5*sin(2 * pi * i / 12)) for i in range(12)]

# Colours of things.
DEFAULT_EDGE_LABEL_COLOUR = 'black'
DEFAULT_EDGE_LABEL_BG_COLOUR = 'white'
MAX_DRAWABLE = 1000  # Maximum weight of a multicurve to draw fully.

Showable = namedtuple('Showable', ['name', 'item'])

def dot(a, b):
    return a[0] * b[0] + a[1] * b[1]

PHI = (1 + sqrt(5)) / 2

def get_colours(num_colours):
    def colour(state):
        hue = (state / PHI) % 1.
        lightness = (50 + random() * 10)/100.
        saturation = (90 + random() * 10)/100.
        r, g, b = hls_to_rgb(hue, lightness, saturation)
        return '#%02x%02x%02x' % (int(r * 255), int(g * 255), int(b * 255))
    
    return [colour(i) for i in range(num_colours)]

class Drawing(object):
    def __init__(self, note, canvas, item, options):
        self.note = note
        self.canvas = canvas
        self.options = options
        self.vertices = []
        self.edges = []
        self.triangles = []
        self.curve_components = []
        self.item = item
        self.lamination = None
        
        if isinstance(item, curver.kernel.MappingClassGroup):
            self.draw_lamination(item.triangulation.empty_lamination())
        if isinstance(item, curver.kernel.Triangulation):
            self.draw_lamination(item.empty_lamination())
        elif isinstance(item, curver.kernel.Lamination):
            self.draw_lamination(item)
        elif isinstance(item, curver.kernel.MappingClass):
            self.draw_lamination(item(item.source_triangulation.as_lamination()))
    
    def get_size(self):
        # Return the size of the canvas.
        # We would like to do:
        # return int(self.canvas.winfo_width()), int(self.canvas.winfo_height())
        # But the canvas may not have been rendered yet so instead we use the note:
        
        return (self.note.winfo_width() - 54, self.note.winfo_height() - 62)
    
    ######################################################################
    
    def create_vertex(self, point):
        assert(isinstance(point, curver.application.Vector2))
        self.vertices.append(curver.application.CanvasVertex(self.canvas, point, self.options))
        return self.vertices[-1]
    
    def create_edge(self, v1, v2, label, colour, create_inverse=False):
        if create_inverse: self.create_edge(v2, v1, ~label, colour)
        self.edges.append(curver.application.CanvasEdge(self.canvas, [v1, v2], label, colour, self.options))
        return self.edges[-1]
    
    def create_triangle(self, e1, e2, e3):
        self.triangles.append(curver.application.CanvasTriangle(self.canvas, [e1, e2, e3], self.options))
        return self.triangles[-1]
    
    def create_curve_component(self, vertices, thin=True, smooth=False):
        assert(all(isinstance(vertex, curver.application.Vector2) for vertex in vertices))
        self.curve_components.append(curver.application.CurveComponent(self.canvas, vertices, self.options, thin, smooth))
        return self.curve_components[-1]
    
    ######################################################################
    
    def translate(self, dx, dy):
        dv = curver.application.Vector2(dx, dy)
        for vertex in self.vertices:
            vertex.vector = vertex.vector + dv
        
        for curve_component in self.curve_components:
            for i in range(len(curve_component.vertices)):
                curve_component.vertices[i] = curve_component.vertices[i] + dv
        
        self.canvas.move('all', dx, dy)
    
    def zoom(self, scale):
        for vertex in self.vertices:
            vertex.vector = scale*vertex.vector
            vertex.update()
        for edge in self.edges:
            edge.update()
        for triangle in self.triangles:
            triangle.update()
        for curve_component in self.curve_components:
            for i in range(len(curve_component.vertices)):
                curve_component.vertices[i] = scale * curve_component.vertices[i]
            curve_component.update()
        self.redraw()
    
    def zoom_centre(self, scale):
        cw, ch = self.get_size()
        self.translate(-cw / 2, -ch / 2)
        self.zoom(scale)
        self.translate(cw / 2, ch / 2)
    
    def zoom_in(self):
        self.zoom_centre(1.05)
    
    def zoom_out(self):
        self.zoom_centre(0.95)
    
    def zoom_to_drawing(self):
        box = self.canvas.bbox('all')
        if box is not None:
            x0, y0, x1, y1 = box
            cw, ch = self.get_size()
            cr = min(cw, ch)
            
            w, h = x1 - x0, y1 - y0
            r = max(w, h)
            
            self.translate(-x0 - w / 2, -y0 - h / 2)
            self.zoom(self.options.zoom_fraction * float(cr) / r)
            self.translate(cw / 2, ch / 2)
    
    ######################################################################
    
    def draw_triangulation(self, triangulation):
        # Get a dual tree.
        dual_tree = triangulation.dual_tree()
        colours = dict((index, None) for index in triangulation.indices)
        outside = [index for index in triangulation.indices if not dual_tree[index]]
        for index, colour in zip(outside, get_colours(len(outside))):
            colours[index] = colour
        components = triangulation.components()
        num_components = len(components)
        cw, ch = self.get_size()
        
        # We will layout the components in a p x q grid.
        # Aim to maximise r := min(cw / p, ch / q) subject to pq >= num_components.
        # There is probably a closed formula for the optimal value of p (and so q).
        p = max(range(1, num_components+1), key=lambda p: min(cw / p, ch / ceil(float(num_components) / p)))
        q = int(ceil(float(num_components) / p))
        
        r = min(cw / p, ch / q) * (1 + self.options.zoom_fraction) / 4
        dx = cw / p
        dy = ch / q
        
        num_used_vertices = 0
        for index, component in enumerate(components):
            n = len(component) // 3  # Number of triangles.
            ngon = n + 2
            
            # Create the vertices.
            for i in range(ngon):
                self.create_vertex(
                    curver.application.Vector2(
                    dx * (index % p) + dx / 2 + r * sin(2 * pi * (i + 0.5) / ngon),
                    dy * int(index / p) + dy / 2 + r * cos(2 * pi * (i + 0.5) / ngon)
                    ))
            
            def num_descendants(edge_label):
                ''' Return the number of triangles that can be reached in the dual tree starting at the given edge_label. '''
                
                corner = triangulation.corner_lookup[edge_label]
                left = (1 + sum(num_descendants(~(corner.labels[2])))) if dual_tree[corner.indices[2]] else 0
                right = (1 + sum(num_descendants(~(corner.labels[1])))) if dual_tree[corner.indices[1]] else 0
                
                return left, right
            
            initial_edge_index = min(edge.index for edge in component if not dual_tree[edge.index])
            to_extend = [(num_used_vertices, num_used_vertices+1, initial_edge_index)]
            # Hmmm, need to be more careful here to ensure that we correctly orient the edges.
            self.create_edge(self.vertices[num_used_vertices+1], self.vertices[num_used_vertices+0], initial_edge_index, colours[initial_edge_index])
            while to_extend:
                source_vertex, target_vertex, label = to_extend.pop()
                left, right = num_descendants(label)
                far_vertex = target_vertex + left + 1
                corner = triangulation.corner_lookup[label]
                
                if corner[2].sign() == +1:
                    self.create_edge(self.vertices[far_vertex], self.vertices[target_vertex], corner[2].label, colours[corner[2].index], left > 0)
                else:
                    self.create_edge(self.vertices[target_vertex], self.vertices[far_vertex], corner[2].label, colours[corner[2].index], left > 0)
                if corner[1].sign() == +1:
                    self.create_edge(self.vertices[source_vertex], self.vertices[far_vertex], corner[1].label, colours[corner[1].index], right > 0)
                else:
                    self.create_edge(self.vertices[far_vertex], self.vertices[source_vertex], corner[1].label, colours[corner[1].index], right > 0)
                
                if left > 0:
                    to_extend.append((far_vertex, target_vertex, ~(corner[2].label)))
                
                if right > 0:
                    to_extend.append((source_vertex, far_vertex, ~(corner[1].label)))
            num_used_vertices = len(self.vertices)
        
        self.edges = sorted(self.edges, key=lambda e: ((0 if e.label >= 0 else 1), e.label))  # So now self.edges[label] is the edge with label label.
        
        for triangle in triangulation:
            self.create_triangle(self.edges[triangle[0].label], self.edges[triangle[1].label], self.edges[triangle[2].label])
    
    def draw_lamination(self, lamination):
        self.draw_triangulation(lamination.triangulation)  # This starts with self.initialise().
        
        vb = self.options.vertex_buffer  # We are going to use this a lot.
        master = float(max(abs(weight) for weight in lamination))
        
        if master > MAX_DRAWABLE:
            for triangle in self.triangles:
                weights = [max(lamination(edge.label), 0) for edge in triangle.edges]
                dual_weights = [lamination.dual_weight(edge.label) for edge in triangle.edges]
                parallel_arcs = [max(-lamination(edge.label), 0) for edge in triangle.edges]
                parallel_weights = [weight // 2 + (weight % 2 if edge.label >= 0 else 0) for edge, weight in zip(triangle.edges, parallel_arcs)]
                for i in range(3):
                    # Dual arcs.
                    a = triangle[i-1] - triangle[i]
                    b = triangle[i-2] - triangle[i]
                    
                    if dual_weights[i] > 0:
                        # We first do the edge to the left of the vertex.
                        # Correction factor to take into account the weight on this edge.
                        s_a = (1 - 2*vb) * weights[i-2] / master
                        # The fractions of the distance of the two points on this edge.
                        scale_a = (1 - s_a) / 2
                        scale_a2 = scale_a + s_a * dual_weights[i] / weights[i-2]
                        
                        # Now repeat for the other edge of the triangle.
                        s_b = (1 - 2*vb) * weights[i-1] / master
                        scale_b = (1 - s_b) / 2
                        scale_b2 = scale_b + s_b * dual_weights[i] / weights[i-1]
                        
                        S1, P1, Q1, E1 = curver.application.interpolate(triangle[i-1].vector, triangle[i].vector, triangle[i-2].vector, scale_a, scale_b)
                        S2, P2, Q2, E2 = curver.application.interpolate(triangle[i-1].vector, triangle[i].vector, triangle[i-2].vector, scale_a2, scale_b2)
                        self.create_curve_component([S1, S1, P1, Q1, E1, E1, E2, E2, Q2, P2, S2, S2, S1, S1], thin=False)
                    elif dual_weights[i] < 0: # Terminal arc.
                        s_0 = (1 - 2*vb) * weights[i] / master
                        
                        scale_a = (1 - s_0) / 2 + s_0 * dual_weights[i-2] / weights[i]
                        scale_a2 = scale_a + s_0 * (-dual_weights[i]) / weights[i]
                        
                        S1, P1, Q1, E1 = curver.application.interpolate(triangle[i-2].vector, triangle[i-1].vector, triangle[i].vector, scale_a, 1.0)
                        S2, P2, Q2, E2 = curver.application.interpolate(triangle[i-2].vector, triangle[i-1].vector, triangle[i].vector, scale_a2, 1.0)
                        self.create_curve_component([S1, S1, P1, E1, E1, P2, S2, S2, S1, S1], thin=False)
                    else:  # dual_weights[i] == 0:  # Nothing to draw.
                        pass
                    
                    # Parallel arcs.
                    if parallel_weights[i]:
                        S, O, E = triangle[i-2].vector, triangle[i].vector, triangle[i-1].vector
                        M = (S + E) / 2.0
                        centroid = (S + O + E) / 3.0
                        P = M + 1.0 * (centroid - M) * (parallel_weights[i]+1) / master
                        self.create_curve_component([S, S, P, E, E, S, S], thin=False)
        else:  # Draw everything. Caution, this is is VERY slow (O(n) not O(log(n))) so we only do it when the weight is low.
            for triangle in self.triangles:
                weights = [max(lamination(edge.label), 0) for edge in triangle.edges]
                dual_weights = [lamination.dual_weight(edge.label) for edge in triangle.edges]
                parallel_arcs = [max(-lamination(edge.label), 0) for edge in triangle.edges]
                parallel_weights = [weight // 2 + (weight % 2 if edge.label >= 0 else 0) for edge, weight in zip(triangle.edges, parallel_arcs)]
                for i in range(3):  # Dual arcs:
                    if dual_weights[i] > 0:
                        s_a = (1 - 2*vb) * weights[i-2] / master
                        s_b = (1 - 2*vb) * weights[i-1] / master
                        for j in range(dual_weights[i]):
                            scale_a = 0.5 if weights[i-2] == 1 else (1 - s_a) / 2 + s_a * j / (weights[i-2] - 1)
                            scale_b = 0.5 if weights[i-1] == 1 else (1 - s_b) / 2 + s_b * j / (weights[i-1] - 1)
                            
                            S, P, Q, E = curver.application.interpolate(triangle[i-1].vector, triangle[i].vector, triangle[i-2].vector, scale_a, scale_b)
                            self.create_curve_component([S, P, Q, E])
                    elif dual_weights[i] < 0: # Terminal arc.
                        s_0 = (1 - 2*vb) * weights[i] / master
                        for j in range(-dual_weights[i]):
                            scale_a = 0.5 if weights[i] == 1 else (1 - s_0) / 2 + s_0 * dual_weights[i-1] / (weights[i] - 1) + s_0 * j / (weights[i] - 1)
                            
                            S, P, Q, E = curver.application.interpolate(triangle[i-2].vector, triangle[i-1].vector, triangle[i].vector, scale_a, 1.0)
                            self.create_curve_component([S, P, E])
                    else:  # dual_weights[i] == 0:  # Nothing to draw.
                        pass
                    
                    # Parallel arcs:
                    S, O, E = triangle[i-2].vector, triangle[i].vector, triangle[i-1].vector
                    M = (S + E) / 2.0
                    centroid = (S + O + E) / 3.0
                    for j in range(parallel_weights[i]):
                        P = M + float(j+1) / (parallel_weights[i] + 1) * (centroid - M) * (parallel_weights[i] + 1) / master
                        self.create_curve_component([S, P, E])
        
        self.lamination = lamination
        self.zoom_to_drawing()  # Recheck.
        self.redraw()  # Install options.
    
    def redraw(self):
        self.create_edge_labels()
        
        self.canvas.itemconfig('line', width=self.options.line_size)
        self.canvas.itemconfig('curve', width=self.options.line_size)
        # Only put arrows on the start so arrow heads appear in the middle.
        for edge in self.edges:
            if self.options.show_orientations and (edge.label >= 0 or self.edges[~edge.label].vertices[::-1] != edge.vertices):
                self.canvas.itemconfig(edge.drawn[0], arrow='last', arrowshape=self.options.arrow_shape)
            else:
                self.canvas.itemconfig(edge.drawn[0], arrow='')
        
        for edge in self.edges:
            edge.hide(not self.options.show_internals and self.edges[~edge.label].vertices[::-1] == edge.vertices)
        self.canvas.tag_raise('polygon')
        self.canvas.tag_raise('line')
        self.canvas.tag_raise('oval')
        self.canvas.tag_raise('curve')
        self.canvas.tag_raise('label')
        self.canvas.itemconfig('curve', smooth=self.options.smooth)
    
    ######################################################################
    
    def destroy_edge_labels(self):
        self.canvas.delete('label')
    
    def create_edge_labels(self):
        self.destroy_edge_labels()  # Remove existing labels.
        
        # How to label each edge.
        if self.options.label_edges == curver.application.options.LABEL_EDGES_INDEX:
            labels = dict((edge.label, edge.index) for edge in self.lamination.triangulation.edges)
        elif self.options.label_edges == curver.application.options.LABEL_EDGES_GEOMETRIC:
            labels = dict((edge.label, self.lamination(edge)) for edge in self.lamination.triangulation.edges)
        elif self.options.label_edges == curver.application.options.LABEL_EDGES_GEOMETRIC_PROJ:
            labels = dict((edge.label, self.lamination(edge)) for edge in self.lamination.triangulation.edges)
            total = float(sum(labels.values()))
            if total != 0:
                # Note the "+ 0" to ensure that -0.0 appears as 0.0.
                labels = dict((label, round(labels[label] / total, 12) + 0) for label in labels)
            else:
                labels = dict((label, 0.0) for label in labels)
        elif self.options.label_edges == curver.application.options.LABEL_EDGES_NONE:
            labels = dict((edge.label, '') for edge in self.lamination.triangulation.edges)
        else:
            raise ValueError()
        
        for edge in self.edges:
            if edge.label >= 0 or self.edges[~edge.label].vertices[::-1] != edge.vertices:
                # We start by creating a nice background for the label. This ensures
                # that it is always readable, even when on top of a lamination.
                # To do this we first draw this label in a different colour with
                # slightly different offsets. This creates a nice 'bubble' effect
                # rather than having to draw a large bounding box.
                for offset in OFFSETS:
                    self.canvas.create_text(
                        [a+x for a, x in zip(edge.centre().to_tuple(), offset)],
                        text=labels[edge.label],
                        tag='label',
                        font=self.options.canvas_font,
                        fill=DEFAULT_EDGE_LABEL_BG_COLOUR)
                
                self.canvas.create_text(edge.centre().to_tuple(),
                    text=labels[edge.label],
                    tag='label',
                    font=self.options.canvas_font,
                    fill=DEFAULT_EDGE_LABEL_COLOUR)


class CurverApplication(object):
    def __init__(self, parent, items):
        self.parent = parent
        self.options = curver.application.Options(self)
        
        s = TTK.Style()
        s.configure('TNotebook', tabposition='n')
        
        self.note = TTK.Notebook(self.parent)
        self.note.pack(fill='both', expand=True, padx=2, pady=2)
        self.note.enable_traversal()
        
        if isinstance(items, (list, tuple)) and len(items) == 1 and isinstance(items[0], (list, dict)):  # [list / dict].
            items = items[0]
        
        if isinstance(items, dict):  # dict.
            items = [Showable(key, item) for key, item in sorted(dict(items).items(), key=lambda x: x[0])]
        try:  # list of pairs.
            items = [Showable(key, item) for key, item in items]
        except (TypeError, ValueError):  # list.
            items = [Showable(name, item) for name, item in curver.kernel.utilities.name_objects(items)]
        self.items = items
        
        self.canvases = [TK.Canvas(self.note, bg='#dcecff')  for i in range(len(self.items))]
        for canvas, item in zip(self.canvases, self.items):
            self.note.add(canvas, text=item.name)
        
        self.parent.wait_visibility(self.note)  # Make sure we get the right sizes of things.
        self.drawings = [Drawing(self.note, canvas, item.item, self.options) for canvas, item in zip(self.canvases, self.items)]
        
        # Create the menus.
        # Make sure to start the Lamination and Mapping class menus disabled.
        self.menubar = TK.Menu(self.parent)
        
        self.filemenu = TK.Menu(self.menubar, tearoff=0)
        self.filemenu.add_command(label='Export image...', command=self.export_image)
        self.filemenu.add_separator()
        self.filemenu.add_command(label='Exit', command=self.quit, accelerator=COMMAND['close'])
        self.menubar.add_cascade(label='File', menu=self.filemenu)
        
        ##########################################
        self.settingsmenu = TK.Menu(self.menubar, tearoff=0)
        
        self.sizemenu = TK.Menu(self.menubar, tearoff=0)
        self.sizemenu.add_radiobutton(label='Small', var=self.options.size_var, value=curver.application.options.SIZE_SMALL)
        self.sizemenu.add_radiobutton(label='Medium', var=self.options.size_var, value=curver.application.options.SIZE_MEDIUM)
        self.sizemenu.add_radiobutton(label='Large', var=self.options.size_var, value=curver.application.options.SIZE_LARGE)
        # self.sizemenu.add_radiobutton(label='Extra large', var=self.options.size_var, value=curver.application.options.SIZE_XLARGE)
        
        self.edgelabelmenu = TK.Menu(self.menubar, tearoff=0)
        self.edgelabelmenu.add_radiobutton(label=curver.application.options.LABEL_EDGES_NONE, var=self.options.label_edges_var)
        self.edgelabelmenu.add_radiobutton(label=curver.application.options.LABEL_EDGES_INDEX, var=self.options.label_edges_var)
        self.edgelabelmenu.add_radiobutton(label=curver.application.options.LABEL_EDGES_GEOMETRIC, var=self.options.label_edges_var)
        self.edgelabelmenu.add_radiobutton(label=curver.application.options.LABEL_EDGES_GEOMETRIC_PROJ, var=self.options.label_edges_var)
        
        self.zoommenu = TK.Menu(self.menubar, tearoff=0)
        self.zoommenu.add_command(label='Zoom in', command=self.zoom_in, accelerator='+')
        self.zoommenu.add_command(label='Zoom out', command=self.zoom_out, accelerator='-')
        self.zoommenu.add_command(label='Zoom to drawing', command=self.zoom_to_drawing, accelerator='0')
        
        self.settingsmenu.add_cascade(label='Sizes', menu=self.sizemenu)
        self.settingsmenu.add_cascade(label='Edge label', menu=self.edgelabelmenu)
        self.settingsmenu.add_cascade(label='Zoom', menu=self.zoommenu)
        self.settingsmenu.add_checkbutton(label='Smooth', var=self.options.smooth_var)
        self.settingsmenu.add_checkbutton(label='Show internal edges', var=self.options.show_internals_var)
        self.settingsmenu.add_checkbutton(label='Show edge orientations', var=self.options.show_orientations_var)
        
        self.menubar.add_cascade(label='Settings', menu=self.settingsmenu)
        
        self.helpmenu = TK.Menu(self.menubar, tearoff=0)
        self.helpmenu.add_command(label='Help', command=self.show_help, accelerator='F1')
        self.helpmenu.add_separator()
        self.helpmenu.add_command(label='About', command=self.show_about)
        
        self.menubar.add_cascade(label='Help', menu=self.helpmenu)
        self.parent.config(menu=self.menubar)
        
        self.parent.bind(COMMAND_KEY['close'], lambda event: self.quit())
        self.parent.bind('<Key>', self.parent_key_press)
        
        self.parent.protocol('WM_DELETE_WINDOW', self.quit)
    
    def current_drawing(self):
        return self.drawings[self.note.index('current')]
    
    def redraw(self):
        for drawing in self.drawings:
            drawing.redraw()
    
    def zoom_in(self):
        self.current_drawing().zoom_in()
    
    def zoom_out(self):
        self.current_drawing().zoom_out()
    
    def zoom_to_drawing(self):
        self.current_drawing().zoom_to_drawing()
    
    def translate(self, dx, dy):
        self.current_drawing().translate(dx, dy)
    
    def parent_key_press(self, event):
        key = event.keysym
        if key == 'F1':
            self.show_help()
        elif key == 'equal' or key == 'plus':
            self.zoom_in()
        elif key == 'minus' or key == 'underscore':
            self.zoom_out()
        elif key == '0':
            self.zoom_to_drawing()
        elif key == 'Up':
            self.translate(0, 5)
        elif key == 'Down':
            self.translate(0, -5)
        elif key == 'Left':
            self.translate(5, 0)
        elif key == 'Right':
            self.translate(-5, 0)
    
    def export_image(self):
        path = tkFileDialog.asksaveasfilename(defaultextension='.ps', filetypes=[('postscript files', '.ps'), ('all files', '.*')], title='Export Image')
        if path:
            try:
                self.current_drawing().canvas.postscript(file=path, colormode='color')
            except IOError:
                tkMessageBox.showwarning('Export Error', 'Could not open: %s' % path)
    
    def quit(self):
        # Apparantly there are some problems with comboboxes, see:
        #  http://stackoverflow.com/questions/15448914/python-tkinter-ttk-combobox-throws-exception-on-quit
        self.parent.eval('::ttk::CancelRepeat')
        self.parent.destroy()
        self.parent.quit()
    
    def show_help(self):
        pass
        # curver.doc.open_documentation()
    
    def show_about(self):
        tkMessageBox.showinfo('About', 'curver (Version %s).\nCopyright (c) Mark Bell 2017.' % curver.__version__)

def start(*items):
    root = TK.Tk()
    root.title('curver')
    root.minsize(300, 300)
    root.geometry('700x500')
    # Set the icon.
    # Make sure to get the right path if we are in a cx_Freeze compiled executable.
    # See: http://cx-freeze.readthedocs.org/en/latest/faq.html#using-data-files
    datadir = os.path.dirname(sys.executable if getattr(sys, 'frozen', False) else __file__)
    icon_path = os.path.join(datadir, 'icon', 'icon.gif')
    img = TK.PhotoImage(file=icon_path)
    try:
        root.tk.call('wm', 'iconphoto', root._w, img)
    except TK.TclError:
        # Give up if we can't set the icon for some reason.
        # This seems to be a problem if you start curver within SnapPy.
        pass
    curver_application = CurverApplication(root, items)
    root.mainloop()

if __name__ == '__main__':
    start()





''' A module for representing more advanced ways of changing triangulations. '''

import curver
from curver.kernel.moves import Move  # Special import needed for subclassing.

class Twist(Move):
    ''' This represents the effect of twisting a short curve.
    
    This format allows us to efficiently perform powers of twists. '''
    def __init__(self, curve, power):
        super(Twist, self).__init__(curve.triangulation, curve.triangulation)
        
        assert(isinstance(curve, curver.kernel.Curve))
        assert(curve.is_short() and not curve.is_peripheral())
        
        self.curve = curve
        self.power = power
        
        a = self.curve.parallel()
        # Theorem: 3*num_tripods is the right number of flips to do in the isolating case.
        # Proof: TODO.
        num_flips = max(self.curve.weight() - 2, 1)  # = 3*num_tripods if self.curve.is_isolating() else 1.
        
        twist = self.curve.triangulation.id_encoding()
        for _ in range(num_flips):
            twist = twist.target_triangulation.encode_flip(twist.target_triangulation.corner_lookup[a.label][2]) * twist
        twist = twist.target_triangulation.find_isometry(twist.source_triangulation, {a.label: a.label}).encode() * twist
        
        self.encoding = twist
    
    def __str__(self):
        return 'Twist^%d_%s ' % (self.power, self.curve)
    def __reduce__(self):
        return (self.__class__, (self.curve, self.power))
    def package(self):
        return (self.curve.parallel().label, self.power)
    def __eq__(self, other):
        if isinstance(other, Twist):
            return self.source_triangulation == other.source_triangulation and self.target_triangulation == other.target_triangulation and self.curve == other.curve and self.power == other.power
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    
    def apply_lamination(self, lamination):
        intersection = self.curve.intersection(lamination)
        if intersection == 0:  # Disjoint twists have no effect.
            return lamination
        
        # Naive way would be to do:
        # for i in range(self.power):
        #     lamination = self.encoding(lamination)
        # return lamination
        # But we can be cleverer and perform this calculation in O(log(self.power)) instead.
        
        power = self.power
        while power:
            slope = self.curve.slope(lamination)
            if power > 0:  # Right twist (increases slope).
                if 1 < slope:  #pylint: disable=misplaced-comparison-constant
                    geometric = [w + power * intersection * c for w, c in zip(lamination, self.curve)]
                    power_applied = power
                elif -1 <= slope <= 1:  # Dangerous region.
                    geometric = self.encoding(lamination).geometric
                    power_applied = 1
                else:  # if slope < -1:
                    steps = min(power, -(slope.numerator // slope.denominator) - 1)  # ceil(-slope) - 1.
                    assert(steps >= 0)  # Sanity.
                    geometric = [w - steps * intersection * c for w, c in zip(lamination, self.curve)]
                    power_applied = steps
            else:  # power < 0:  # Left twist (decreases slope).
                if slope < -1:
                    geometric = [w + -power * intersection * c for w, c in zip(lamination, self.curve)]
                    power_applied = power
                elif -1 <= slope <= 1:  # Dangerous region.
                    geometric = self.encoding.inverse()(lamination).geometric
                    power_applied = -1
                else:  # 1 < slope:
                    steps = min(-power, -(-slope.numerator // slope.denominator) - 1)  # ceil(slope) - 1.
                    assert(steps >= 0)  # Sanity.
                    geometric = [w - steps * intersection * c for w, c in zip(lamination, self.curve)]
                    power_applied = -steps
            new_lamination = lamination.__class__(self.target_triangulation, geometric)  # Avoids promote.
            # assert(new_lamination == (self.encoding**power_applied)(lamination))
            # assert(-1 <= slope <= 1 or self.curve.slope(new_lamination) == slope + power_applied)
            lamination = new_lamination
            power = power - power_applied
        
        return lamination
    
    def apply_homology(self, homology_class):
        # I don't think we even need this case.
        if self.curve.is_isolating():  # Isolating ==> separating, so no effect on homology.
            return homology_class
        
        a = self.curve.parallel()
        _, b, e = self.source_triangulation.corner_lookup[a.label]
        
        algebraic = list(homology_class)
        algebraic[a.index] -= a.sign() * self.power * (homology_class(e) - homology_class(b))
        
        return curver.kernel.HomologyClass(self.target_triangulation, algebraic)
    
    def inverse(self):
        return Twist(self.curve, -self.power)

class HalfTwist(Move):
    ''' This represents the effect of half-twisting a short arc.
    
    This format allows us to efficiently perform powers of twists. '''
    def __init__(self, arc, power):
        super(HalfTwist, self).__init__(arc.triangulation, arc.triangulation)
        
        assert(isinstance(arc, curver.kernel.Arc))
        assert(arc.is_short())
        
        self.arc = arc
        self.power = power
        
        conjugator = arc.triangulation.id_encoding()
        # We need to get to a really good configuration, one where self.arc is not just short
        # but where valence(self.arc.initial_vertex) == 1.
        
        edge = self.arc.parallel()
        # Reverse the orientation if the valence of the other end is less.
        # This reduces the number of flips needed to reach a really good configuration.
        if len(arc.triangulation.vertex_lookup[edge.label]) > len(arc.triangulation.vertex_lookup[~edge.label]):
            edge = ~edge
        
        # Since self.arc is short it is an edge of the triangulation so we just keep moving
        # edges away from this edge's initial vertex to get to a really good triangulation.
        while len(conjugator.target_triangulation.vertex_lookup[edge.label]) > 1:  # valence(initial vertex) > 1.
            flip = conjugator.target_triangulation.encode_flip(conjugator.target_triangulation.corner_lookup[edge.label][2])
            conjugator = flip * conjugator
        
        # We can now perform the half twist. To do this we move all the edges back across to the other vertex.
        # Again, we keep moving edges away from this edge's terminal vertex.
        # TODO: 4) Prove this always works.
        # NOTE: William Worden checked that this works for genus <= 20.
        half_twist = conjugator.target_triangulation.id_encoding()  # valence(terminal vertex) > 1.
        while len(half_twist.target_triangulation.vertex_lookup[~edge.label]) > 1:
            flip = half_twist.target_triangulation.encode_flip(half_twist.target_triangulation.corner_lookup[~edge.label][2])
            half_twist = flip * half_twist
        
        # No close up to complete the half twist. Use the isometry that inverts this edge.
        half_twist = half_twist.target_triangulation.find_isometry(half_twist.source_triangulation, {edge.label: ~edge.label}).encode() * half_twist
        
        self.encoding = conjugator.inverse() * half_twist * conjugator
        
        # We handle large powers by replacing (T^1/2_self)^2 with T_boundary, which includes acceleration.
        # We handle small powers separately to increase performance.
        if abs(self.power) <= 1:
            self.encoding_power = self.encoding**self.power
        elif self.power % 2 == 0:
            self.encoding_power = self.arc.boundary().encode_twist(self.power // 2)
        else:  # self.power % 2 == 1:  # Division rounds down so, regardless of power, we need an extra right half-twist.
            self.encoding_power = self.arc.boundary().encode_twist(self.power // 2) * self.encoding
    
    def __str__(self):
        return 'HalfTwist^%d_%s ' % (self.power, self.arc)
    def __reduce__(self):
        return (self.__class__, (self.arc, self.power))
    def package(self):
        return (self.arc.parallel().label, self.power)
    def __eq__(self, other):
        if isinstance(other, HalfTwist):
            return self.source_triangulation == other.source_triangulation and self.target_triangulation == other.target_triangulation and self.arc == other.arc and self.power == other.power
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    
    def apply_lamination(self, lamination):
        return self.encoding_power(lamination)
    
    def apply_homology(self, homology_class):
        return self.encoding_power(homology_class)
    
    def inverse(self):
        return HalfTwist(self.arc, -self.power)





''' A module for the various different errors that can be raised. '''

class AbortError(Exception):
    ''' An exception for aborting computations with.
    
    This is thrown by clicking 'cancel' on a progress box. '''
    
    def __init__(self, message=None):
        super(AbortError, self).__init__()
        self.message = message
    def __str__(self):
        return str(self.message)

class AssumptionError(Exception):
    ''' An exception for when an assumption is false. '''
    
    def __init__(self, message=None):
        super(AssumptionError, self).__init__()
        self.message = message
    def __str__(self):
        return str(self.message)





''' The curver kernel.

Some of the functions and methods have assumptions on them. These are denoted in the docstrings
by "Assumes that ..." meaning that:

    - If the assumptions are met then this function is guaranteed to terminate correctly.
    - If not then a curver.AssumptionError will be raised. '''

from . import arc
from . import crush
from . import curve
from . import curvecomplex
from . import encoding
from . import error
from . import homology
from . import lamination
from . import mappingclassgroup
from . import moves
from . import partition
from . import permutation
from . import traintrack
from . import triangulation
from . import twist
from . import utilities

# Set up shorter names for all of the different classes.
Edge = triangulation.Edge
Triangle = triangulation.Triangle
Triangulation = triangulation.Triangulation
Encoding = encoding.Encoding
MappingClass = encoding.MappingClass
AbortError = error.AbortError
AssumptionError = error.AssumptionError
MappingClassGroup = mappingclassgroup.MappingClassGroup
MCG = MappingClassGroup  # Alias.
HomologyClass = homology.HomologyClass
Lamination = lamination.Lamination
Shortenable = lamination.Shortenable
MultiCurve = curve.MultiCurve
Curve = curve.Curve
MultiArc = arc.MultiArc
Arc = arc.Arc
Move = moves.Move
Isometry = moves.Isometry
EdgeFlip = moves.EdgeFlip
TrainTrack = traintrack.TrainTrack
Crush = crush.Crush
Lift = crush.Lift
CurveComplex = curvecomplex.CurveComplex
Twist = twist.Twist
HalfTwist = twist.HalfTwist
CurvePartitionGraph = partition.CurvePartitionGraph
Permutation = permutation.Permutation

norm = triangulation.norm
UnionFind = utilities.UnionFind
memoize = utilities.memoize  # Unused?
matrix_vector = utilities.matrix_vector

# Functions that help with construction.
create_triangulation = Triangulation.from_tuple
triangulation_from_sig = Triangulation.from_sig





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies

class TestTwist(unittest.TestCase):
    @given(strategies.curves(), st.integers())
    @settings(max_examples=3)
    def test_pickle(self, curve, power):
        twist_i = curve.encode_twist(power)
        
        self.assertEqual(twist_i, pickle.loads(pickle.dumps(twist_i)))
    
    @given(strategies.curves(), st.integers(), st.integers())
    @settings(max_examples=5)
    def test_powers(self, curve, power1, power2):
        twist_i = curve.encode_twist(power1)
        twist_j = curve.encode_twist(power2)
        twist_ij = curve.encode_twist(power1 + power2)
        twist_neg_i = curve.encode_twist(-power1)
        
        self.assertEqual(twist_i * twist_j, twist_j * twist_i)  # Commute.
        self.assertEqual(twist_i * twist_j, twist_ij)  # Additive.
        self.assertEqual(twist_neg_i, ~twist_i)  # Inverse.





''' The curver GUI application. '''

from . import main
from . import pieces
from . import options
from . import vector

# Set up shorter names for all of the different classes and some common constructors.
start = main.start
Options = options.Options
CurverApplication = main.CurverApplication
CanvasVertex = pieces.CanvasVertex
CanvasEdge = pieces.CanvasEdge
CanvasTriangle = pieces.CanvasTriangle
CurveComponent = pieces.CurveComponent
Vector2 = vector.Vector2

interpolate = pieces.interpolate





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies

class TestTriangulation(unittest.TestCase):
    @given(strategies.triangulations())
    def test_pickle(self, triangulation):
        self.assertEqual(triangulation, pickle.loads(pickle.dumps(triangulation)))
    
    @given(st.data())
    def test_hash(self, data):
        triangulation1 = data.draw(strategies.triangulations())
        triangulation2 = data.draw(strategies.triangulations())
        self.assertTrue(hash(triangulation1) != hash(triangulation2) or triangulation1 == triangulation2)
    
    @given(st.data())
    def test_flip(self, data):
        triangulation = data.draw(strategies.triangulations())
        edge = data.draw(st.sampled_from(triangulation.edges))
        self.assertEqual(triangulation.surface(), triangulation.flip_edge(edge).surface())
    
    @given(strategies.triangulations())
    def test_isometry(self, triangulation):
        self.assertTrue(triangulation.is_isometric_to(triangulation))
        identity = triangulation.id_isometry()
        isometries = triangulation.self_isometries()
        self.assertIn(identity, isometries)
    
    @given(st.data())
    def test_relabel(self, data):
        triangulation = data.draw(strategies.triangulations())
        label_map = [i if data.draw(st.booleans()) else ~i for i in data.draw(st.permutations(range(triangulation.zeta)))]
        
        T = triangulation.relabel_edges(label_map)
        self.assertTrue(triangulation.is_isometric_to(T))
        self.assertTrue(any(all(isom.label_map[i] == label_map[i] for i in range(triangulation.zeta)) for isom in triangulation.isometries_to(T)))
    
    @given(strategies.triangulations())
    def test_sig(self, triangulation):
        self.assertEqual(triangulation, curver.triangulation_from_sig(triangulation.sig()))
    
    @given(strategies.triangulations())
    def test_homology(self, triangulation):
        self.assertEqual(len(triangulation.homology_basis()), 1 - triangulation.euler_characteristic)





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies

class TestArc(unittest.TestCase):
    @given(strategies.arcs())
    @settings(max_examples=25)
    def test_boundary_intersection(self, arc):
        boundary = arc.boundary()
        self.assertEqual(arc.intersection(boundary), 0)
    
    @given(strategies.arcs().filter(lambda a: a.connects_distinct_vertices()))
    @settings(max_examples=50)
    def test_halftwist(self, arc):
        self.assertEqual(arc.boundary().encode_twist(), arc.encode_halftwist()**2)





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
from collections import Counter

# TRIANGULATIONS = {(g, p): curver.load(g, p).triangulation.sig() for g in range(10) for p in range(1, 10) if 6*g + 3*p - 6 >= 3}
TRIANGULATIONS = {
    (0, 3): '3_p',
    (0, 4): '6_JDky1',
    (0, 5): '9_ZFO4OuIV',
    (0, 6): 'c_ZlgeM906o-354',
    (0, 7): 'f_ZBhisze7ZhPen+Kdl1',
    (0, 8): 'i_ZBCol3oodJti4fabgfrLdu1',
    (0, 9): 'l_ZBmFdZ5TmXJTHbErcDHjdC12vbv4',
    (1, 1): '3_o',
    (1, 2): '6_WKSv',
    (1, 3): '9_q3AtQIBp',
    (1, 4): 'c_qfj5RWuFLb802',
    (1, 5): 'f_qLjT1kC5ZTnGyy1-H',
    (1, 6): 'i_qLbeF0856iXdcySItPVI7O',
    (1, 7): 'l_qLH58G-B-C5woL+MsXnS96SB0Ur2',
    (1, 8): 'o_qLH3BJ5cZdEYqFJK8US2y-FobflUMuJXh',
    (1, 9): 'r_qLHzvY4m6fubS+XBc6Fe3q2oydqacuUFP8knNo4',
    (2, 1): '9_qBFcQEN4',
    (2, 2): 'c_q5YHALgCWTyw',
    (2, 3): 'f_q5sG87f308pRczx4k',
    (2, 4): 'i_q5s8zbvSopJhvzCSeTFSHs',
    (2, 5): 'l_q5sEtshJlZcP3VxnSg50en+0fYA1',
    (2, 6): 'o_q5sEjz8JquEV7VEKDk7V1jA2MXn78nNzc',
    (2, 7): 'r_q5sEXh18WxhR6p+c1t3FqBE5ADBr5HuhrOHDBf3',
    (2, 8): 'u_q5sEXJrSMGIad7plECHCF-ZDpcWLW83ZkNe4eiHHmFlD1',
    (2, 9): 'x_q5sEXd3XbSQm+-kXyrXCVYKYFcTz6Z5jHLHM+udBWiyxoM5Zbt1',
    (3, 1): 'f_0mGaGVParMxLYnCS3',
    (3, 2): 'i_0mGG3b5sZ48hMGl+XMxQf7',
    (3, 3): 'l_0mGGHziLfC85djn-maLAMR4st0K',
    (3, 4): 'o_0mGGHv87rZuOX6acFcu2Pxuu+DqmvfRb7',
    (3, 5): 'r_0mGGH-frv5VxbMIDZJg+PfsvsGkD+T4V2s3Up62',
    (3, 6): 'u_0mGGH-P8pozBmiNXbK0fZpo0ncTEiu04QdNjBn9ZJgn81',
    (3, 7): 'x_0mGGH-3ktQlFDgcyM9fkSxjeT1cb5E-PwJDw3I-lkIOMDUAcV41',
    (3, 8): 'A_0mGGH-3rGANCo9JE6TmZaKNI6kuZUobdsYwLbLayo8Ub3hD75-FeycbL1',
    (3, 9): 'D_0mGGH-3XUmszRC-f0Kyqkw4UM8-vbAhPgaq-V4Gb6SjpdRrqXpP6MLNexfnVuA4',
    (4, 1): 'l_0wtlMpgmgKA+k2ObfDEOlf+tdih1',
    (4, 2): 'o_0wtlMFe+inn8rzo8rBgSeeSJdPKrw1OFa',
    (4, 3): 'r_0wtlMFCguCalYSPETZQ9G4DvnpCtAr6eMZFLPR2',
    (4, 4): 'u_0wtlMF+WbZ81Ll+SKPg5sOpjSoG6gQWjbTLQZ38oPavs1',
    (4, 5): 'x_0wtlMF+eOhAEjOyEX96St+C4dYcwSbwozSel1VJ0du3wvIPKJk1',
    (4, 6): 'A_0wtlMF+uWL8zHWMw5In5jPZGKqe3qImPInE7fKCNV2uUrxwodSD9H7X42',
    (4, 7): 'D_0wtlMF+uW2PslyYLdtjcGMFYloqw8elhB75LSzd7FlORRH0vo8y8+7Bsxnr0lm5',
    (4, 8): 'G_0wtlMF+uKiqZQ4wW85xwYGeuWoKctihKeCaJXLhpKtRsAj00sS960b23XP0t67ospXIOl',
    (4, 9): 'J_0wtlMF+uKu-uOzqzZ+7kU2Lq4I3BEeaaVa57q2C0lhFS7iV1YBCH7IbI-L3HePt7iIGBe5WgA282',
    (5, 1): 'r_0wUXPZwUD9uA4K-fmxT93jrye5mW+6QUPZVGQU1',
    (5, 2): 'u_0wUXPZw7sxaMvf+HM6p0m50javFCeVyOYYsOtrBCvcH21',
    (5, 3): 'x_0wUXPZwTnJKfEIppmxHddHc+ER8yCqQ445v0d2Q101uZCcV6u01',
    (5, 4): 'A_0wUXPZwTTUvsjvktbsTgIgjJ7aqCDyJNKbtky0Ajvrz4SWEQ+5CjlC9F1',
    (5, 5): 'D_0wUXPZwTHMJ9w2zmjkrNnNtK7FwyVbF-mL4UHfta7eN54N2lZekGdxwvKMt1Om4',
    (5, 6): 'G_0wUXPZwTHYiyz+sdTFf3khZz195gt99PeEiX+MLv9mUa4O4yfOqo3b--xuvog9dP2fO9i',
    (5, 7): 'J_0wUXPZwTH0cWeOU-yEvWsxsiJOElP-Vml7Bi2JyY24itGFBv1Z0bif-hsK1c9zL-QETRIcKg9pP1',
    (5, 8): 'M_0wUXPZwTH0cRiVJPBRoeCnlyjorpiQI8coG+UQtINlFoP-I6sOedoZGzA8cPUK3lFgQIaV7vSIaoDajr1h',
    (5, 9): 'P_0wUXPZwTH0sVLX2dmkM054OZAYTbfU7OhNURAF5-lmsXhmepRzsMjqrme2wP+y2oFw16Ud-04jQ7uCBbFXMocr5F3',
    }

SIGNATURES = [TRIANGULATIONS[key] for key in sorted(TRIANGULATIONS)]

@st.composite
def triangulations(draw):
    sig = draw(st.sampled_from(SIGNATURES))
    return curver.triangulation_from_sig(sig)

MCGS = [curver.load(g, p) for g in range(0, 5) for p in range(1, 5) if 6*g + 3*p - 6 >= 3]

@st.composite
def mcgs(draw):
    return draw(st.sampled_from(MCGS))

@st.composite
def mapping_classes(draw, mcg=None):
    if mcg is None: mcg = draw(mcgs())
    return mcg(draw(st.integers(min_value=0, max_value=10)))

@st.composite
def encodings(draw, triangulation=None):
    if triangulation is None: triangulation = draw(triangulations())
    rev_sequence = [triangulation.id_isometry()]
    num_flips = draw(st.integers(min_value=0, max_value=20))
    for _ in range(num_flips):
        T = rev_sequence[-1].target_triangulation
        if draw(st.sampled_from([0, 0, 0, 0, 0, 1])) == 0:
            edge = draw(st.sampled_from([edge for edge in T.edges if T.is_flippable(edge)]))
            move = T.encode_flip(edge)[0]
        else:
            move = T.encode_relabel_edges([i if draw(st.booleans()) else ~i for i in draw(st.permutations(range(T.zeta)))])[0]
        rev_sequence.append(move)
    
    return curver.kernel.Encoding(rev_sequence[::-1])

@st.composite
def homology_classes(draw, triangulation=None):
    if triangulation is None: triangulation = draw(triangulations())
    algebraic = [draw(st.integers()) for _ in range(triangulation.zeta)]
    return curver.kernel.HomologyClass(triangulation, algebraic)

@st.composite
def multiarcs(draw, triangulation=None, require_non_empty_boundary=False):
    if triangulation is None: triangulation = draw(triangulations())
    
    geometric = [0] * triangulation.zeta
    
    available_indices = set(triangulation.indices)
    num_arcs = draw(st.integers(min_value=1, max_value=len(available_indices)))
    for _ in range(num_arcs):
        index = draw(st.sampled_from(sorted(available_indices)))
        available_indices.remove(index)
        geometric[index] = draw(st.integers(max_value=-1))
    
    return triangulation.lamination(geometric)

@st.composite
def arcs(draw, triangulation=None):
    if triangulation is None: triangulation = draw(triangulations())
    
    return draw(multiarcs(triangulation)).peek_component()

@st.composite
def multicurves(draw, triangulation=None):
    if triangulation is None: triangulation = draw(triangulations())
    
    # Build a special type of multiarc with non-empty boundary.
    geometric = [0] * triangulation.zeta
    
    indices = set()
    available_indices = set(triangulation.indices)
    
    tree = triangulation.dual_tree()
    tree = set([index for index in triangulation.indices if tree[index]])
    available_indices = available_indices - tree
    
    if draw(st.booleans()) and not any(g == 0 for (g, v) in triangulation.surface()):  # merge vertices:
        classes = curver.kernel.utilities.UnionFind(triangulation.vertices)
        for index in sorted(available_indices):
            a, b = triangulation.edge_arc(index).vertices()
            if classes(a) != classes(b):
                classes.union(a, b)
                indices.add(index)
                available_indices.remove(index)
    
    num_arcs = draw(st.integers(min_value=0 if indices else 1, max_value=len(available_indices)-1))
    for _ in range(num_arcs):
        index = draw(st.sampled_from(sorted(available_indices)))
        available_indices.remove(index)
        indices.add(index)
    
    # Set weights on these edges.
    for index in indices:
        geometric[index] = draw(st.integers(max_value=-1))
    
    multiarc = triangulation.lamination(geometric)
    boundary = multiarc.boundary()
    assert(not boundary.is_empty())
    return boundary

@st.composite
def curves(draw, triangulation=None):
    if triangulation is None: triangulation = draw(triangulations())
    
    multicurve = draw(multicurves(triangulation))
    curve = multicurve.peek_component()
    return curve

@st.composite
def laminations(draw, triangulation=None):
    return draw(st.one_of(multicurves(triangulation), multiarcs(triangulation)))

@st.composite
def permutations(draw, N=None):
    if N is None: N = draw(st.integers(min_value=1, max_value=10))
    return curver.kernel.Permutation(draw(st.permutations(range(N))))


class TestStrategiesHealth(unittest.TestCase):
    @given(triangulations())
    def test_triangulations(self, triangulation):
        self.assertIsInstance(triangulation, curver.kernel.Triangulation)
    
    @given(mcgs())
    def test_mcgs(self, mcg):
        self.assertIsInstance(mcg, curver.kernel.MappingClassGroup)
    
    @given(mapping_classes())
    def test_mapping_classes(self, h):
        self.assertIsInstance(h, curver.kernel.MappingClass)
    
    @given(encodings())
    def test_encodings(self, encoding):
        self.assertIsInstance(encoding, curver.kernel.Encoding)
    
    @given(homology_classes())
    def test_homology_classes(self, hc):
        self.assertIsInstance(hc, curver.kernel.HomologyClass)
    
    @given(arcs())
    def test_arcs(self, arc):
        self.assertIsInstance(arc, curver.kernel.Arc)
    
    @given(multiarcs())
    def test_multiarcs(self, multiarc):
        self.assertIsInstance(multiarc, curver.kernel.MultiArc)
    
    @given(curves())
    def test_curves(self, curve):
        self.assertIsInstance(curve, curver.kernel.Curve)
    
    @given(multicurves())
    def test_multicurves(self, multicurve):
        self.assertIsInstance(multicurve, curver.kernel.MultiCurve)
    
    @pytest.mark.skip('Incomplete')
    @given(laminations())
    def test_laminations(self, lamination):
        self.assertIsInstance(lamination, curver.kernel.Lamination)
    
    @given(permutations())
    def test_permutations(self, perm):
        self.assertIsInstance(perm, curver.kernel.Permutation)


if __name__ == '__main__':
    unittest.main()





try:
    import Tkinter as TK
    import tkFont as TK_FONT
except ImportError:  # Python 3.
    import tkinter as TK
    import tkinter.font as TK_FONT

try:
    import ttk as TTK
except ImportError:  # Python 3.
    from tkinter import ttk as TTK

LABEL_EDGES_NONE = 'None'
LABEL_EDGES_INDEX = 'Index'
LABEL_EDGES_GEOMETRIC = 'Geometric'
LABEL_EDGES_GEOMETRIC_PROJ = 'Projective geometric'
SIZE_SMALL, SIZE_MEDIUM, SIZE_LARGE = 0, 1, 2
# SIZE_XLARGE = 3

class Options(object):
    def __init__(self, parent):
        self.parent = parent
        self.application_font = TK_FONT.Font(family='TkDefaultFont', size=10)
        self.canvas_font = TK_FONT.Font(family='TkDefaultFont', size=10)
        
        self.show_internals_var = TK.BooleanVar(value=False)
        self.show_orientations_var = TK.BooleanVar(value=False)
        self.straight_laminations_var = TK.BooleanVar(value=False)
        self.smooth_var = TK.BooleanVar(value=True)
        self.label_edges_var = TK.StringVar(value=LABEL_EDGES_NONE)
        self.size_var = TK.IntVar(value=SIZE_SMALL)
        
        self.show_internals = False
        self.show_orientations = False
        self.straight_laminations = False
        self.label_edges = LABEL_EDGES_NONE
        self.smooth = True
        self.line_size = 2
        self.dot_size = 3
        self.arrow_shape = (12, 15, 5)
        
        # Set it so that self.update() will be called whenever these variables are changed.
        self.show_internals_var.trace('w', self.update)
        self.show_orientations_var.trace('w', self.update)
        self.straight_laminations_var.trace('w', self.update)
        self.smooth_var.trace('w', self.update)
        self.label_edges_var.trace('w', self.update)
        self.size_var.trace('w', self.update)
        
        # Drawing parameters.
        self.epsilon = 10
        self.float_error = 0.001
        
        self.vertex_buffer = 0.2  # Must be in (0, 0.5)
        self.zoom_fraction = 0.8 # Must be in (0, 1)
    
    def update(self, *args):
        self.show_internals = bool(self.show_internals_var.get())
        self.show_orientations = bool(self.show_orientations_var.get())
        self.straight_laminations = bool(self.straight_laminations_var.get())
        self.smooth = bool(self.smooth_var.get())
        self.label_edges = str(self.label_edges_var.get())
        if self.size_var.get() == SIZE_SMALL:
            self.line_size = 2
            self.dot_size = 3
            self.arrow_shape = (12, 15, 5)
            self.application_font.configure(size=10)
            self.canvas_font.configure(size=10)
        elif self.size_var.get() == SIZE_MEDIUM:
            self.line_size = 4
            self.dot_size = 5
            self.arrow_shape = (16, 20, 6)
            self.application_font.configure(size=11)
            self.canvas_font.configure(size=12)
        elif self.size_var.get() == SIZE_LARGE:
            self.line_size = 6
            self.dot_size = 7
            self.arrow_shape = (24, 30, 9)
            self.application_font.configure(size=12)
            self.canvas_font.configure(size=14)
        
        self.parent.redraw()





''' A module for representing permutations in Sym(N). '''

from bisect import bisect
from itertools import combinations
from math import factorial

import curver

class Permutation(object):
    ''' This represents a permutation on 0, 1, ..., N-1. '''
    def __init__(self, perm):
        self.perm = perm
        assert(set(self) == set(range(len(self))))
    
    def __str__(self):
        return str(self.perm)
    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.perm)
    def __getitem__(self, item):
        return self.perm[item]
    def __call__(self, item):
        return self[item]
    def __iter__(self):
        return iter(self.perm)
    def __len__(self):
        return len(self.perm)
    def __eq__(self, other):
        if isinstance(other, Permutation):
            if len(self) != len(other):
                raise ValueError('Cannot compare permutations defined over different number of elements.')
            
            return self.perm == other.perm
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    def __hash__(self):
        return hash(tuple(self.perm))
    
    def inverse(self):
        ''' Return the inverse of this permutation. '''
        
        return Permutation(sorted(range(len(self)), key=self))
    def __invert__(self):
        return self.inverse()
    
    def order(self):
        ''' Return the order of this permutation. '''
        
        identity = Permutation(list(range(len(self))))
        power = self
        i = 1
        while True:
            if power == identity: return i
            i += 1
            power = power * self
    
    def __mul__(self, other):
        if isinstance(other, Permutation):
            if len(self) != len(other):
                raise ValueError('Cannot compose permutations defined over different number of elements.')
            
            return Permutation([self(other(i)) for i in range(len(self))])
        else:
            return NotImplemented
    
    def __pow__(self, n):
        if n < 0: return (~self)**(-n)
        
        result = Permutation(list(range(len(self))))
        while n:
            if n % 2 == 1:
                result = result * self
                n = n - 1
            self = self * self
            n = n // 2
        return result
    
    @classmethod
    def from_index(cls, N, index):
        ''' Return the permutation in Sym(N) with the given index. '''
        
        P = []
        f = factorial(N)
        symbols = list(range(N))
        while symbols:
            f = f // len(symbols)
            i, index = divmod(index, f)
            P.append(symbols[i])
            symbols = symbols[:i] + symbols[i+1:]
        return cls(P)
    
    def index(self):
        ''' Return the index of this permutation in the (sorted) list of all permutations on this many symbols. '''
        
        symbols = sorted(self.perm)
        index = 0
        for p in self:
            i = bisect(symbols, p) - 1
            index = index * len(symbols) + i
            symbols = symbols[:i] + symbols[i+1:]
        return index
    
    def is_even(self):
        ''' Return whether this permutation is the composition of an even number of transposiions. '''
        
        return sum(1 if a > b else 0 for a, b in combinations(self, r=2) if a > b) % 2 == 0





''' A module for representing more advanced ways of changing triangulations. '''

import curver
from curver.kernel.moves import Move  # Special import needed for subclassing.

class Crush(Move):
    ''' This represents the effect of crushing along a curve. '''
    def __init__(self, source_triangulation, target_triangulation, curve, matrix):
        super(Crush, self).__init__(source_triangulation, target_triangulation)
        
        assert(isinstance(curve, curver.kernel.Curve))
        assert(curve.is_short() and not curve.is_peripheral())
        assert(curve.triangulation == self.source_triangulation)
        
        self.curve = curve
        self.matrix = matrix
    
    def __str__(self):
        return 'Crush ' + str(self.curve)
    def __reduce__(self):
        return (self.__class__, (self.source_triangulation, self.target_triangulation, self.curve, self.matrix))
    def __eq__(self, other):
        if isinstance(other, Crush):
            return self.source_triangulation == other.source_triangulation and self.target_triangulation == other.target_triangulation and self.curve == other.curve and self.matrix == other.matrix
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    
    def apply_lamination(self, lamination):
        geometric = list(lamination)
        
        if self.curve.is_isolating():
            # Get some edges.
            a = self.curve.parallel()
            v = self.curve.triangulation.vertex_lookup[a.label]  # = self.triangulation.vertex_lookup[~a.label].
            _, b, e = self.source_triangulation.corner_lookup[a.label]
            
            v_edges = curver.kernel.utilities.cyclic_slice(v, a, ~a)  # The set of edges that come out of v from a round to ~a.
            around_v = min(max(lamination.side_weight(edge), 0) for edge in v_edges)
            out_v = sum(max(-lamination.side_weight(edge), 0) for edge in v_edges) + sum(max(-lamination(edge), 0) for edge in v_edges[1:])
            # around_v > 0 ==> out_v == 0; out_v > 0 ==> around_v == 0.
            
            # TODO: 4) Add comments explaining what is going on here and how the different cases work.
            # We work by manipulating the dual weights around v.
            twisting = min(max(lamination.side_weight(edge) - around_v, 0) for edge in v_edges[1:-1])
            sides = dict()
            for edge in v_edges:
                if lamination.side_weight(edge) < 0:
                    sides[edge] = lamination.side_weight(edge)
                else:  # lamination.side_weight(edge) >= 0:
                    if edge == a or edge == b:
                        sides[edge] = lamination.side_weight(edge) - around_v
                    elif edge == e:
                        sides[edge] = lamination.side_weight(edge) - 2*twisting - around_v
                    else:
                        sides[edge] = lamination.side_weight(edge) - twisting - around_v
            parallels = dict((edge.index, max(-lamination(edge), 0)) for edge in v_edges)
            
            # Tighten to the left.
            drop = max(sides[a], 0) + max(-sides[b], 0)
            for edge in v_edges[1:-1]:
                x, y, z = lamination.triangulation.corner_lookup[edge.label]
                if sides[x] >= 0 and sides[y] >= 0 and sides[z] >= 0:
                    if drop <= sides[x]:
                        sides[x] = sides[x] - drop
                    else:  # sides[x] < drop.
                        sides[x], sides[y], drop = sides[x] - drop, sides[y] + sides[x] - drop, sides[x]
                elif sides[x] < 0:
                    sides[x], sides[y], drop = sides[x] - drop, sides[y] - drop, 0
                elif sides[y] < 0:
                    sides[x] = sides[x] - drop
                else:  # sides[z] < 0.
                    if drop <= sides[x]:
                        sides[x] = sides[x] - drop
                    elif sides[x] < drop <= sides[x] - sides[z]:
                        parallels[z.index] = parallels[z.index] + (drop - sides[x])
                        sides[x], sides[z], drop = 0, sides[z] + (drop - sides[x]), sides[x]
                    else:  # sides[x] - sides[z] < drop:
                        parallels[z.index] = parallels[z.index] - sides[z]
                        sides[x], sides[y], sides[z], drop = sides[x] - sides[z] - drop, sides[y] - (drop - sides[x] + sides[z]), 0, sides[x]
                
                if drop == 0: break  # Stop early.
            
            # Tighten to the right.
            drop = max(-sides[a], 0) + max(sides[b], 0)
            for edge in v_edges[-2:0:-1]:
                x, y, z = lamination.triangulation.corner_lookup[edge.label]
                if sides[x] >= 0 and sides[y] >= 0 and sides[z] >= 0:
                    if drop <= sides[x]:
                        sides[x] = sides[x] - drop
                    else:  # sides[x] < drop.
                        sides[x], sides[z], drop = sides[x] - drop, sides[z] + sides[x] - drop, sides[x]
                elif sides[x] < 0:
                    sides[x], sides[z], drop = sides[x] - drop, sides[z] - drop, 0
                elif sides[y] < 0:
                    if drop <= sides[x]:
                        sides[x] = sides[x] - drop
                    elif sides[x] < drop <= sides[x] - sides[y]:
                        parallels[x.index] = parallels[x.index] + (drop - sides[x])
                        sides[x], sides[y], drop = 0, sides[y] + (drop - sides[x]), sides[x]
                    else:  # sides[x] - sides[y] < drop:
                        parallels[x.index] = parallels[x.index] - sides[y]
                        sides[x], sides[y], sides[z], drop = sides[x] - sides[y] - drop, 0, sides[z] - (drop - sides[x] + sides[y]), sides[x]
                else:  # sides[z] < 0.
                    sides[x] = sides[x] - drop
                
                if drop == 0: break  # Stop early.
            
            # Now rebuild the intersection.
            for edge in v_edges:
                if edge not in (a, b, e, ~b, ~e):
                    x, y, z = lamination.triangulation.corner_lookup[edge.label]
                    if parallels[edge.index] > 0:
                        geometric[edge.index] = -parallels[x.index]
                    else:
                        geometric[edge.index] = max(sides[x], 0) + max(sides[y], 0) + max(-sides[z], 0)
                        
                        # Sanity check:
                        x2, y2, z2 = lamination.triangulation.corner_lookup[~edge.label]
                        assert(geometric[edge.index] == max(sides[x2], 0) + max(sides[y2], 0) + max(-sides[z2], 0))
            
            # We have to rebuild the ~e edge separately since it now pairs with ~b.
            x, y, z = lamination.triangulation.corner_lookup[~e.label]
            if parallels[e.index] + parallels[b.index] > 0:
                geometric[e.index] = -parallels[e.index] - parallels[b.index]
            else:
                geometric[e.index] = max(sides[x], 0) + max(sides[y], 0) + max(-sides[z], 0)
                
                # Sanity check:
                x2, y2, z2 = lamination.triangulation.corner_lookup[~b.label]
                assert(geometric[e.index] == max(sides[x2], 0) + max(sides[y2], 0) + max(-sides[z2], 0))
            
            # And finally the b edge, which is now paired with e.
            geometric[b.index] = around_v - out_v  # Same trick as below.
        else:  # self.curve is non-isolating.
            # Get some edges.
            a = self.curve.parallel()
            _, b, e = self.source_triangulation.corner_lookup[a.label]
            _, c, d = self.source_triangulation.corner_lookup[~e.label]
            
            u = self.source_triangulation.vertex_lookup[a.label]  # = self.triangulation.vertex_lookup[~a.label].
            u_edges = curver.kernel.utilities.cyclic_slice(u, a, ~a)
            around_u = min(max(lamination.side_weight(edge), 0) for edge in u_edges)  # The number of components that go through a, around u and then back through ~a.
            out_u = sum(max(-lamination.side_weight(edge), 0) for edge in u_edges) + sum(max(-lamination(edge), 0) for edge in u_edges[1:])  # The number of arcs that come out of u (from a around to ~a).
            # Since around_u > 0 ==> out_u == 0 & out_u > 0 ==> around_u == 0, this is equivalent to around_u if around_u > 0 else -out_u
            geometric[b.index] = around_u - out_u
            
            v = self.source_triangulation.vertex_lookup[c.label]  # = self.triangulation.vertex_lookup[~c.label].
            v_edges = curver.kernel.utilities.cyclic_slice(v, c, ~c)
            around_v = min(max(lamination.side_weight(edge), 0) for edge in v_edges)
            out_v = sum(max(-lamination.side_weight(edge), 0) for edge in v_edges) + sum(max(-lamination(edge), 0) for edge in v_edges[1:])
            geometric[e.index] = around_v - out_v  # Same trick.
        
        return self.target_triangulation(geometric)  # Have to promote.
    
    def apply_homology(self, homology_class):
        return NotImplemented  # I don't think we ever need this.
    
    def inverse(self):
        return Lift(self.target_triangulation, self.source_triangulation, self.curve, self.matrix)

class Lift(Move):
    ''' This represents the inverse of crushing along a curve. '''
    def __init__(self, source_triangulation, target_triangulation, curve, matrix):
        super(Lift, self).__init__(source_triangulation, target_triangulation)
        
        assert(isinstance(curve, curver.kernel.Curve))
        assert(curve.triangulation == self.target_triangulation)
        
        self.curve = curve
        self.matrix = matrix
    
    def __str__(self):
        return 'Lift ' + str(self.curve)
    def __reduce__(self):
        return (self.__class__, (self.source_triangulation, self.target_triangulation, self.curve, self.matrix))
    def __eq__(self, other):
        if isinstance(other, Lift):
            return self.source_triangulation == other.source_triangulation and self.target_triangulation == other.target_triangulation and self.curve == other.curve and self.matrix == other.matrix
        else:
            return NotImplemented
    def __ne__(self, other):
        return not (self == other)
    
    def apply_lamination(self, lamination):
        # Really should check that the dual weights around a vertex are all non-negative.
        geometric = curver.kernel.matrix_vector(self.matrix, lamination.geometric)
        return lamination.__class__(self.target_triangulation, geometric)  # Avoid promote since the lift has to be the same type as the given lamination.
    
    def apply_homology(self, homology_class):
        return NotImplemented  # I don't think we ever need this.
    
    def inverse(self):
        return Crush(self.target_triangulation, self.source_triangulation, self.curve, self.matrix)





from hypothesis import given, settings
import hypothesis.strategies as st
import pickle
import pytest
import unittest

import curver
import strategies

class TestHomologyClass(unittest.TestCase):
    @given(strategies.homology_classes())
    def test_pickle(self, hc):
        self.assertEqual(hc, pickle.loads(pickle.dumps(hc)))
    
    @given(st.data())
    def test_hash(self, data):
        hc1 = data.draw(strategies.homology_classes())
        hc2 = data.draw(strategies.homology_classes(hc1.triangulation))
        self.assertTrue(hash(hc1) != hash(hc2) or hc1 == hc2)
    
    @given(st.data())
    def test_canonical(self, data):
        hc1 = data.draw(strategies.homology_classes())
        hc2 = data.draw(strategies.homology_classes(hc1.triangulation))
        self.assertEqual(hc1.canonical() + hc2.canonical(), (hc1 + hc2).canonical())
        self.assertEqual(hc1.canonical() - hc2.canonical(), (hc1 - hc2).canonical())
        self.assertEqual(-(hc1.canonical()), (-hc1).canonical())
    
    @given(st.data())
    def test_image(self, data):
        hc1 = data.draw(strategies.homology_classes())
        hc2 = data.draw(strategies.homology_classes(hc1.triangulation))
        h = data.draw(strategies.encodings(hc1.triangulation))
        self.assertEqual(h(hc1 + hc2), h(hc1) + h(hc2))
    
    @given(st.data())
    def test_abelian(self, data):
        T = data.draw(strategies.triangulations())
        hcs = data.draw(st.lists(elements=strategies.homology_classes(T), min_size=1))
        self.assertEqual(sum(hcs), sum(data.draw(st.permutations(hcs))))
    
    @given(st.data())
    def test_orientation(self, data):
        hc = data.draw(strategies.homology_classes())
        edge = data.draw(st.sampled_from(hc.triangulation.edges))
        self.assertEqual(hc(edge), -hc(~edge))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''The setup script.'''

from setuptools import setup, find_packages

def read(file_name):
    with open(file_name) as open_file:
        return open_file.read()

requirements = [
    'networkx',
    'numpy',
]

test_requirements = [
    'pytest',
    'hypothesis',
    'pylint',
    'pytest',
    'pytest-cov'
]

setup(
    name='curver',
    version='0.1.0',
    description='For calculations in the curve complex',
    long_description='See http://curver.readthedocs.io for the full README, LICENCE and documentation.',
    author='Mark Bell',
    author_email='mcbell@illinois.edu',
    url='https://github.com/MarkCBell/curver',
    packages=find_packages(),
    package_data = {
        'curver.application': ['icon/*'],
        },
    include_package_data=True,
    install_requires=requirements,
    license='GNU General Public License v3',
    zip_safe=False,
    keywords='curver',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Mathematics',
        ],
    test_suite='tests',
    tests_require=test_requirements,
)




''' A module of useful, generic functions; including input and output formatting. '''

from functools import wraps
from itertools import product
from string import ascii_lowercase, ascii_uppercase, digits
from collections import defaultdict
import re

import curver

ALPHABET = digits + ascii_lowercase + ascii_uppercase + '+-'

def b64encode(n):
    ''' Return n in base 64. '''
    
    out = []
    while n:
        n, r = divmod(n, 64)
        out.append(ALPHABET[r])
    return ''.join(out)

def b64decode(strn):
    ''' Return the integer with base 64 encoding strn. '''
    
    return sum(ALPHABET.index(c) * 64**i for i, c in enumerate(strn))

def string_generator(n, skip=None):
    ''' Return a list of n usable names, none of which are in skip. '''
    
    assert(isinstance(n, curver.IntegerType))
    assert(skip is None or isinstance(skip, (list, tuple, dict, set)))
    
    skip = set() if skip is None else set(skip)
    if n < 1: return []
    
    alphabet = ascii_lowercase
    results = []
    i = 0
    while True:
        i += 1
        for letters in product(alphabet, repeat=i):
            word = ''.join(letters)
            if word not in skip:
                results.append(word)
            if len(results) >= n:
                return results

def name_objects(objects, skip=None):
    ''' Return a list of pairs (name, object). '''
    
    assert(isinstance(objects, (list, tuple)))
    
    return zip(string_generator(len(objects), skip), objects)

class UnionFind(object):
    ''' A fast union--find data structure. Given items must be hashable. '''
    def __init__(self, items):
        self.parent = dict((item, item) for item in items)
        self.rank = dict((item, 0) for item in items)
    def __iter__(self):
        ''' Iterate through the groups of self. '''
        groups = defaultdict(list)
        for item in self.parent:
            groups[self(item)].append(item)
        return iter(groups.values())
    def __len__(self):
        return sum(1 if self.parent[item] == item else 0 for item in self.parent)
    def __repr__(self):
        return str(self)
    def __str__(self):
        return ', '.join('{' + ', '.join(str(item) for item in g) + '}' for g in self)
    def __call__(self, x):
        ''' Find the root of x. Two items are in the same group iff they have the same root. '''
        root = x
        while self.parent[root] != root:
            root = self.parent[root]
        while self.parent[x] != root:
            x, self.parent[x] = self.parent[x], root
        return root
    def union2(self, x, y):
        ''' Combine the class containing x and the class containing y. '''
        rx, ry = self(x), self(y)
        if self.rank[x] > self.rank[y]:
            self.parent[ry] = rx
        elif self.rank[x] < self.rank[y]:
            self.parent[rx] = ry
        elif rx != ry:
            self.parent[ry] = rx
            self.rank[rx] += 1
    def union(self, *args):
        ''' Combine all of the classes containing the given items. '''
        if len(args) == 1: args = args[0]
        for item in args:
            self.union2(args[0], item)

def memoize(fast=False):
    ''' A decorator that memoizes a method of a class. '''
    def memoizer(function):
        ''' A decorator that memoizes a method of a class and knows if the method is fast. '''
        @wraps(function)
        def caching(self, _is_fast=False):
            ''' The cached version of function.
            
            Note that this docstring will be overwritten with functions docstring by the wraps decorator. '''
            if not hasattr(self, '_cache'):
                self._cache = dict()
            key = function.__name__
            if _is_fast:
                return fast or key in self._cache
            else:
                if key not in self._cache:
                    self._cache[key] = function(self)
                return self._cache[key]
        
        return caching
    return memoizer

def cyclic_slice(L, x, y):
    ''' Return the sublist of L from x (inclusive) to y (exclusive).
    
    L may be cyclically permuted if needed. '''
    i = L.index(x)
    L = L[i:] + L[:i]  # x is now L[0].
    j = L.index(y)
    return L[:j]

def maximum(iterable, key=lambda x: x, upper_bound=None):
    ''' Return the maximum of iterable but terminate early when given an upper_bound. '''

    def helper():
        ''' A generator that yeilds elements from iterable up to and including one such that key(item) >= upper_bound. '''
        for item in iterable:
            yield item
            if upper_bound is not None and key(item) >= upper_bound: return

    return max(helper(), key=key)

def alphanum_key(strn):
    ''' Return a list of string and number chunks from a string. '''
    
    blocks = []
    for chunk in re.split('([0-9]+)', strn):
        try:
            blocks.append(int(strn))
        except ValueError:
            blocks.append(strn)
    
    return blocks

def matrix_vector(M, v):
    ''' Return M*v where v is interpreted as a column vector. '''
    
    return (v * M.T).tolist()[0]





''' A simple starting point for curver. '''

import curver

def main():
    ''' Describe how to start curver. '''
    
    print('curver %s' % curver.__version__)
    print('Some basic commands:')
    print('  > python -m curver.app      # To start the curver GUI.')
    print('Or start Python and import curver directly.')

if __name__ == '__main__':
    main()

